#!/usr/bin/env python3


import json
from datetime import datetime

from .hook import Hook
from ..result import TestResult, CaseResult, StepResult


class StdLogHook(Hook):
    def __init__(self, args):
        super().__init__(args)

    def on_exit(self, res: TestResult):
        StdLogHook.log(res.name, "test", res.status, res.elapse, {
            "casePass": res.case_pass,
            "caseFail": res.case_fail,
            "caseSkip": res.case_skip,
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_test_end(self, res: TestResult):
        StdLogHook.log(res.name, "subTest", res.status, res.elapse, {
            "casePass": res.case_pass,
            "caseFail": res.case_fail,
            "caseSkip": res.case_skip,
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_set_up_end(self, res: CaseResult):
        StdLogHook.log(res.name, "setUp", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_case_end(self, res: CaseResult):
        StdLogHook.log(res.name, "case", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_tear_down_end(self, res: CaseResult):
        StdLogHook.log(res.name, "tearDown", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_step_end(self, res: StepResult):
        if res.is_skip:
            status = "skip"
        elif res.is_pass:
            status = "pass"
        else:
            status = "fail"
        StdLogHook.log(res.name, "step", status, res.elapse, {
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    @staticmethod
    def log(name, type_, status, elapse, detail):
        print(json.dumps({
            "time": datetime.utcnow().replace(microsecond=0).isoformat(),
            "type": type_,
            "status": status,
            "elapseMs": int(elapse.total_seconds() * 1000),
            "detail": detail
        }))



#!/usr/bin/env python3


import itertools
import functools


def generate_req(d, p="!"):
    if isinstance(d, list):
        return itertools.product(*[generate_req(v) for v in d])
    if isinstance(d, dict):
        return [
            dict(i)
            for i in itertools.product(*[
                [
                    (k[1:], i) for i in v
                ]
                if k.startswith(p) else
                [
                    *[(k, i) for i in generate_req(v)]
                ]
                for k, v in sorted(d.items())
            ])
        ]
    return d,


def generate_res(d, n, p="!"):
    if not d:
        return itertools.repeat(d, n)
    if isinstance(d, list):
        return zip(*[generate_res(v, n) for v in d])
    if isinstance(d, dict):
        return [
            dict(i)
            for i in zip(*[
                [(k[1:], i) for i in v]
                if k.startswith(p) else
                [(k, i) for i in generate_res(v, n)]
                for k, v in sorted(d.items())
            ])
        ]
    return itertools.repeat(d, n)


def calculate_num(d, p="!"):
    if isinstance(d, list):
        return functools.reduce(lambda x, y: x * y, [calculate_num(v) for v in d], 1)
    if isinstance(d, dict):
        return functools.reduce(lambda x, y: x * y, [len(v) if k.startswith(p) else calculate_num(v) for k, v in d.items()], 1)
    return 1


def grouper(iterable, n):
    if n == 0:
        yield iterable

    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk



#!/usr/bin/env python3


from .include import *
from .exec_func import exec_with_res, exec_shell


def render(req, case=None, var=None, x=None, peval="#", pexec="%", pshell="$"):
    if isinstance(req, dict):
        res = {}
        for key, val in req.items():
            if key.startswith(peval):
                res[key[len(peval):]] = eval(val)
            elif key.startswith(pexec):
                res[key[len(pexec):]] = exec_with_res(val, case=case, var=var, x=x)
            elif key.startswith(pshell):
                res[key[len(pshell):]] = exec_shell(val)
            else:
                res[key] = render(req[key], case=case, var=var, x=x)
        return res
    if isinstance(req, list):
        res = []
        for val in req:
            res.append(render(val, case=case, var=var, x=x))
        return res
    return req



#!/usr/bin/env python3


import json
import pymongo
from bson.objectid import ObjectId

from ..util import merge, REQUIRED
from .driver import Driver


class MongoDriver(Driver):
    def __init__(self, args):
        args = merge(args, {
            "host": "localhost",
            "port": 27017,
            "database": REQUIRED,
        })

        self.client = pymongo.MongoClient(
            host=args["host"],
            port=args["port"],
        )
        self.database = self.client.get_database(args["database"])

    def name(self, req):
        return req["cmd"]

    def do(self, req):
        req = merge(req, {
            "cmd": REQUIRED
        })

        do_map = {
            "insertOne": self.insert_one,
            "findOne": self.find_one,
            "updateOne": self.update_one,
            "deleteOne": self.delete_one,
        }
        if req["cmd"] not in do_map:
            raise Exception("unsupported action [{}]".format(req["cmd"]))

        try:
            return do_map[req["cmd"]](req)
        except Exception as e:
            raise e

    def insert_one(self, req):
        req = merge(req, {
            "collection": REQUIRED,
            "document": REQUIRED,
        })
        collection = self.database.get_collection(req["collection"])
        res = collection.insert_one(req["document"])
        return {"_id": str(res.inserted_id)}

    def find_one(self, req):
        req = merge(req, {
            "collection": REQUIRED,
            "filter": {},
            "_id": "",
        })
        collection = self.database.get_collection(req["collection"])
        if req["_id"]:
            filter_ = {"_id": ObjectId(req["_id"])}
        else:
            filter_ = req["filter"]
        res = collection.find_one(filter_)
        return json.loads(json.dumps(dict(res), default=lambda x: str(x)))

    def update_one(self, req):
        req = merge(req, {
            "collection": REQUIRED,
            "filter": {},
            "_id": "",
            "update": {},
            "upsert": False,
        })
        collection = self.database.get_collection(req["collection"])
        if req["_id"]:
            filter_ = {"_id": ObjectId(req["_id"])}
        else:
            filter_ = req["filter"]

        res = collection.update_one(filter_, {"$set": flat(req["update"])}, upsert=req["upsert"])

        return {
            "acknowledged": res.acknowledged,
            "matchedCount": res.matched_count,
            "modifiedCount": res.modified_count,
        }

    def delete_one(self, req):
        req = merge(req, {
            "collection": REQUIRED,
            "filter": {},
            "_id": "",
        })
        collection = self.database.get_collection(req["collection"])
        if req["_id"]:
            filter_ = {"_id": ObjectId(req["_id"])}
        else:
            filter_ = req["filter"]
        res = collection.delete_one(filter_)
        return {
            "acknowledged": res.acknowledged,
            "deleteCount": res.deleted_count,
        }


def flat(d):
    res = {}
    _flat_recursive([], d, res)
    return res


def _flat_recursive(k, d, res):
    for key in d:
        if isinstance(d[key], dict):
            _flat_recursive(k + [key], d[key], res)
        else:
            res[".".join(k + [key])] = d[key]



#!/usr/bin/env python3


import datetime
import json
from jinja2 import Environment, BaseLoader
import markdown
from types import SimpleNamespace

from ..result import TestResult, SubStepResult
from ..util import merge, REQUIRED
from .reporter import Reporter
from .format_step_res import format_step_res


_report_tpl = """<!DOCTYPE html>
<html lang="zh-cmn-Hans">
<head>
    <title>{{ res.name }} {{ i18n.title.report }}</title>
    <meta charset="UTF-8">
    <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-1BmE4kWBq78iYhFldvKuhfTAU6auU8tT94WrHftjDbrCEXSU1oBoqyl2QvZ6jIW3" crossorigin="anonymous">
    <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js" integrity="sha384-ka7Sk0Gln4gmtz2MlQnikT1wXgYsOg+OMhuP+IlRH9sENBO0LRn5q+8nbTov4+1p" crossorigin="anonymous"></script>
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.1/font/bootstrap-icons.css">
    <script src="https://code.jquery.com/jquery-3.6.0.slim.min.js" integrity="sha256-u7e5khyithlIdTpu22PHhENmPcRdFiHRjhAuHcs05RI=" crossorigin="anonymous"></script>
    {{ customize.font.style }}
    <style>
        body {
            font-family: {{ customize.font.body }};
        }
        pre, code {
            font-family: {{ customize.font.code }};
        }
    </style>
    <script>
        function copyToClipboard(elementId) {
            var aux = document.createElement("textarea");
            aux.value = document.getElementById(elementId).textContent;
            aux.style.height = "0";
            aux.style.overflow = "hidden";
            aux.style.position = "fixed";
            document.body.appendChild(aux);
            aux.select();
            document.execCommand("copy");
            document.body.removeChild(aux);
        }

        function showTestPass(parentID) {
            $('#' + parentID + ' .test.test-skip').hide()
            $('#' + parentID + ' .test.test-fail').hide()
            $('#' + parentID + ' .test.test-pass').show()
        }

        function showTestSkip(parentID) {
            $('#' + parentID + ' .test.test-pass').hide()
            $('#' + parentID + ' .test.test-fail').hide()
            $('#' + parentID + ' .test.test-skip').show()
        }

        function showTestFail(parentID) {
            $('#' + parentID + ' .test.test-pass').hide()
            $('#' + parentID + ' .test.test-skip').hide()
            $('#' + parentID + ' .test.test-fail').show()
        }

        function showCasePass(parentID) {
            $('#' + parentID + ' .test.case-skip').hide()
            $('#' + parentID + ' .test.case-fail').hide()
            $('#' + parentID + ' .test.case-pass').show()
            $('#' + parentID + ' .case.skip').hide()
            $('#' + parentID + ' .case.fail').hide()
            $('#' + parentID + ' .case.pass').show()
        }

        function showCaseSkip(parentID) {
            $('#' + parentID + ' .test.case-pass').hide()
            $('#' + parentID + ' .test.case-fail').hide()
            $('#' + parentID + ' .test.case-skip').show()
            $('#' + parentID + ' .case.pass').hide()
            $('#' + parentID + ' .case.fail').hide()
            $('#' + parentID + ' .case.skip').show()
        }

        function showCaseFail(parentID) {
            $('#' + parentID + ' .test.case-pass').hide()
            $('#' + parentID + ' .test.case-skip').hide()
            $('#' + parentID + ' .test.case-fail').show()
            $('#' + parentID + ' .case.pass').hide()
            $('#' + parentID + ' .case.skip').hide()
            $('#' + parentID + ' .case.fail').show()
        }

        var testToggleStatus = {}
        var caseToggleStatus = {}

        function showAllTest(parentID) {
            if (testToggleStatus[parentID]) {
                $('#' + parentID + ' .test.test-pass').show()
                $('#' + parentID + ' .test.test-skip').show()
                $('#' + parentID + ' .test.test-fail').show()
                testToggleStatus[parentID] = false
            } else {
                $('#' + parentID + ' .test.test-pass').hide()
                $('#' + parentID + ' .test.test-skip').hide()
                $('#' + parentID + ' .test.test-fail').hide()
                testToggleStatus[parentID] = true
            }
        }

        function showAllCase(parentID) {
            if (caseToggleStatus[parentID]) {
                $('#' + parentID + ' .test.test-pass').show()
                $('#' + parentID + ' .test.test-skip').show()
                $('#' + parentID + ' .test.test-fail').show()
                $('#' + parentID + ' .case.pass').show()
                $('#' + parentID + ' .case.skip').show()
                $('#' + parentID + ' .case.fail').show()
                caseToggleStatus[parentID] = false
            } else {
                $('#' + parentID + ' .test.test-pass').hide()
                $('#' + parentID + ' .test.test-skip').hide()
                $('#' + parentID + ' .test.test-fail').hide()
                $('#' + parentID + ' .case.pass').hide()
                $('#' + parentID + ' .case.skip').hide()
                $('#' + parentID + ' .case.fail').hide()
                caseToggleStatus[parentID] = true
            }
        }
    </script>

    {{ customize.extra.head }}
</head>

<body>
    {{ customize.extra.bodyHeader }}
    <div class="container">
        <div class="row justify-content-md-center">
            <div class="col-lg-10 col-md-12">
            {{ render_test(res, "test") }}
            </div>
        </div>
    </div>
    {{ customize.extra.bodyFooter }}
</body>

<script>
    var tooltipTriggerList = [].slice.call(document.querySelectorAll('[data-bs-toggle="tooltip"]'))
    var tooltipList = tooltipTriggerList.map(function (tooltipTriggerEl) {
      return new bootstrap.Tooltip(tooltipTriggerEl)
    })
</script>
</html>
"""

_test_tpl = """
<div class="col-md-12" id={{ name }}>
    {% if res.is_skip %}
    <div class="card my-{{ customize.padding.y }} border-warning bg-warning">
        <h5 class="card-header text-white bg-transparent border-0">{{ i18n.title.test }} {{ res.name }} {{ i18n.status.skip }}</h5>
    {% elif res.is_pass %}
    <div class="card my-{{ customize.padding.y }} border-success">
        <h5 class="card-header text-white bg-success">{{ i18n.title.test }} {{ res.name }} {{ i18n.status.succ }}</h5>
    {% else %}
    <div class="card my-{{ customize.padding.y }} border-danger">
        <h5 class="card-header text-white bg-danger">{{ i18n.title.test }} {{ res.name }} {{ i18n.status.fail }}</h5>
    {% endif %}

        {% if not res.is_skip %}
        <div class="card-body">
            <table class="table table-striped">
                <thead>
                    <tr class="text-center">
                        <th>{{ i18n.summary.caseTotal }}</th>
                        <th>{{ i18n.summary.casePass }}</th>
                        <th>{{ i18n.summary.caseSkip }}</th>
                        <th>{{ i18n.summary.caseFail }}</th>
                        <th>{{ i18n.summary.stepPass }}</th>
                        <th>{{ i18n.summary.stepSkip }}</th>
                        <th>{{ i18n.summary.stepFail }}</th>
                        <th>{{ i18n.summary.assertionPass }}</th>
                        <th>{{ i18n.summary.assertionFail }}</th>
                        <th>{{ i18n.summary.elapse }}</th>
                    </tr>
                </thead>
                <tbody>
                    <tr class="text-center">
                        {% if res.case_pass %}
                        <td><span class="badge bg-primary rounded-pill" onclick="showAllCase('{{ name }}')">{{ res.case_pass + res.case_skip + res.case_fail }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.case_pass }}</span></td>
                        {% endif %}

                        {% if res.case_pass %}
                        <td><span class="badge bg-success rounded-pill" onclick="showCasePass('{{ name }}')">{{ res.case_pass }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.case_pass }}</span></td>
                        {% endif %}

                        {% if res.case_skip %}
                        <td><span class="badge bg-warning rounded-pill" onclick="showCaseSkip('{{ name }}')">{{ res.case_skip }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.case_skip }}</span></td>
                        {% endif %}

                        {% if res.case_fail %}
                        <td><span class="badge bg-danger rounded-pill" onclick="showCaseFail('{{ name }}')">{{ res.case_fail }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.case_fail }}</span></td>
                        {% endif %}

                        {% if res.step_pass %}
                        <td><span class="badge bg-success rounded-pill">{{ res.step_pass }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.step_pass }}</span></td>
                        {% endif %}

                        {% if res.step_skip %}
                        <td><span class="badge bg-warning rounded-pill">{{ res.step_skip }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.step_skip }}</span></td>
                        {% endif %}

                        {% if res.step_fail %}
                        <td><span class="badge bg-danger rounded-pill">{{ res.step_fail }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.step_fail }}</span></td>
                        {% endif %}

                        {% if res.assertion_pass %}
                        <td><span class="badge bg-success rounded-pill">{{ res.assertion_pass }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.assertion_pass }}</span></td>
                        {% endif %}

                        {% if res.assertion_fail %}
                        <td><span class="badge bg-danger rounded-pill">{{ res.assertion_fail }}</span></td>
                        {% else %}
                        <td><span class="badge bg-secondary rounded-pill">{{ res.assertion_fail }}</span></td>
                        {% endif %}

                        <td>{{ format_timedelta(res.elapse) }}</td>
                    </tr>
                </tbody>
            </table>
        </div>
        {% endif %}
        
        {# 渲染 Err #}
        {% if res.is_err %}
        <div class="card-header text-white bg-danger"><span class="fw-bolder">{{ i18n.testHeader.err }}</span></div>
        <div class="card-body"><pre>{{ res.err }}</pre></div>
        {% endif %}

        {# 渲染 Description #}
        {% if res.description %}
        <div class="card-header justify-content-between d-flex"><span class="fw-bolder">{{ i18n.testHeader.description }}</span></div>
        <div class="card-body">{{ markdown(res.description) }}</div>
        {% endif %}


        {# 渲染 set_up #}
        {% if res.set_ups %}
        <div class="card-header justify-content-between d-flex">
            <span class="fw-bolder">{{ i18n.testHeader.setUp }}</span>
            <span>
                <span class="badge bg-primary rounded-pill" onclick="showAllCase('{{ name }}-set-up')">{{ res.set_up_pass + res.set_up_fail }}</span>
                {% if res.set_up_pass %}
                <span class="badge bg-success rounded-pill" onclick="showCasePass('{{ name }}-set-up')">{{ res.set_up_pass }}</span>
                {% endif %}
                {% if res.set_up_fail %}
                <span class="badge bg-danger rounded-pill" onclick="showCaseFail('{{ name }}-set-up">{{ res.set_up_fail }}</span>
                {% endif %}
            </span>
        </div>
        <ul class="list-group list-group-flush" id="{{ name }}-set-up">
            {% for case in res.set_ups %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }} case {{ case.status }}">
                {{ render_case(case, '{}-set-up-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}

        {# 渲染 case #}
        {% if res.cases %}
        <div class="card-header justify-content-between d-flex {{ "pass" if res.case_fail == 0 else "fail" }}">
            <span class="fw-bolder">{{ i18n.testHeader.case }}</span>
            <span>
                <span class="badge bg-primary rounded-pill" onclick="showAllCase('{{ name }}-case')">{{ res.curr_case_pass + res.curr_case_skip + res.curr_case_fail }}</span>
                {% if res.curr_case_pass %}
                <span class="badge bg-success rounded-pill" onclick="showCasePass('{{ name }}-case')">{{ res.curr_case_pass }}</span>
                {% endif %}
                {% if res.curr_case_skip %}
                <span class="badge bg-warning rounded-pill" onclick="showCaseSkip('{{ name }}-case')">{{ res.curr_case_skip }}</span>
                {% endif %}
                {% if res.curr_case_fail %}
                <span class="badge bg-danger rounded-pill" onclick="showCaseFail('{{ name }}-case')">{{ res.curr_case_fail }}</span>
                {% endif %}
            </span>
        </div>
        <ul class="list-group list-group-flush" id="{{ name }}-case">
            {% for case in res.cases %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }} case {{ case.status }}">
                {{ render_case(case, '{}-case-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}

        {# 渲染 subtest #}
        {% if res.sub_tests %}
        <div class="card-header justify-content-between d-flex">
            <span class="fw-bolder">{{ i18n.testHeader.subTest }}</span>
            <span>
                <span class="badge bg-primary rounded-pill" onclick="showAllTest('{{ name }}-subtest')">{{ res.sub_test_pass + res.sub_test_skip + res.sub_test_fail }}</span>
                {% if res.sub_test_pass %}
                <span class="badge bg-success rounded-pill" onclick="showTestPass('{{ name }}-subtest')">{{ res.sub_test_pass }}</span>
                {% endif %}
                {% if res.sub_test_skip %}
                <span class="badge bg-warning rounded-pill" onclick="showTestSkip('{{ name }}-subtest')">{{ res.sub_test_skip }}</span>
                {% endif %}
                {% if res.sub_test_fail %}
                <span class="badge bg-danger rounded-pill" onclick="showTestFail('{{ name }}-subtest')">{{ res.sub_test_fail }}</span>
                {% endif %}
            </span>
        </div>
        <ul class="list-group list-group-flush" id="{{ name }}-subtest">
            {% for sub_test in res.sub_tests %}
            <li class="
                list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }} test
                test-{{ sub_test.status }}
                {% if sub_test.case_pass %}case-pass{% endif %}
                {% if sub_test.case_skip %}case-skip{% endif %}
                {% if sub_test.case_fail %}case-fail{% endif %}
            ">
                {{ render_test(sub_test, '{}-subtest-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}

        {# 渲染 tear_down #}
        {% if res.tear_downs %}
        <div class="card-header justify-content-between d-flex">
            <span class="fw-bolder">{{ i18n.testHeader.tearDown }}</span>
            <span>
                <span class="badge bg-primary rounded-pill" onclick="showAllCase('{{ name }}-tear-down')">{{ res.tear_down_pass + res.tear_down_fail }}</span>
                {% if res.tear_down_pass %}
                <span class="badge bg-success rounded-pill" onclick="showCasePass('{{ name }}-tear-down')">{{ res.tear_down_pass }}</span>
                {% endif %}
                {% if res.tear_down_fail %}
                <span class="badge bg-danger rounded-pill" onclick="showCaseFail('{{ name }}-tear-down')">{{ res.tear_down_fail }}</span>
                {% endif %}
            </span>
        </div>
        <ul class="list-group list-group-flush" id="{{ name }}-tear-down">
            {% for case in res.tear_downs %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }} case {{ case.status }}">
                {{ render_case(case, '{}-tear-down-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}
    </div>
</div>
"""

_case_tpl = """
{% if case.is_skip %}
<a class="card-title btn d-flex justify-content-between align-items-center">
    {{ case.name }}
    <span class="badge bg-warning rounded-pill">skip</span>
</a>
{% elif case.is_pass %}
<a class="card-title btn d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    {{ case.name }}
    <span>{{ format_timedelta(case.elapse) }}</span>
</a>
{% else %}
<a class="card-title text-white bg-danger btn btn-danger d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    {{ case.name }}
    <span>{{ format_timedelta(case.elapse) }}</span>
</a>
{% endif %}
<div class="collapse card" id="{{ name }}">
    {% if case.is_pass %}
    <div class="card border-success">
    {% else %}
    <div class="card border-danger">
    {% endif %}
        {# Description #}
        {% if case.description %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.description }}</span></div>
        <div class="card-body">{{ markdown(case.description) }}</div>
        {% endif %}
        
        {% if case.command %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.command }}</span></div>
        <div class="card-body">
            <div class="float-end">
                <button type="button" class="btn btn-sm py-0" onclick="copyToClipboard('{{ name }}-command')"
                    data-bs-toggle="tooltip" data-bs-placement="top" title="{{ i18n.toolTips.copy }}">
                    <i class="bi-clipboard"></i>
                </button>
            </div>
            <span id="{{ name }}-command">{{ case.command }}</span>
        </div>        
        {% endif %}

        {# BeforeCaseStep #}
        {% if case.before_case_steps %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.beforeCaseStep }}</span></div>
        <ul class="list-group list-group-flush">
            {% for step in case.before_case_steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_step(step, '{}-before-case-step-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}
        
        {# PreStep #}
        {% if case.pre_steps %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.preStep }}</span></div>
        <ul class="list-group list-group-flush">
            {% for step in case.pre_steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_step(step, '{}-pre-step-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}
        
        {# Step #}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.step }}</span></div>
        <ul class="list-group list-group-flush">
            {% for step in case.steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_step(step, '{}-step-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        
        {# PostStep #}
        {% if case.pre_steps %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.postStep }}</span></div>
        <ul class="list-group list-group-flush">
            {% for step in case.post_steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_step(step, '{}-post-step-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}

        {# AfterCaseStep #}
        {% if case.after_case_steps %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.caseHeader.afterCaseStep }}</span></div>
        <ul class="list-group list-group-flush">
            {% for step in case.after_case_steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_step(step, '{}-after-case-step-{}'.format(name, loop.index0)) }}
            </li>
            {% endfor %}
        </ul>
        {% endif %}
    </div>
</div>
"""

_step_tpl = """
{% if step.is_skip %}
<a class="card-title btn d-flex justify-content-between align-items-center">
    {{ step.ctx }}
    <span class="badge bg-warning rounded-pill">skip</span>
</a>
{% elif step.is_pass %}
<a class="card-title btn d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    {{ step.ctx }}.{{ step.name }}
    <span>{{ format_timedelta(step.elapse) }}</span>
</a>
{% else %}
<a class="card-title text-white bg-danger btn btn-danger d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    {{ step.ctx }}.{{ step.name }}
    <span>{{ format_timedelta(step.elapse) }}</span>
</a>
{% endif %}

<div class="collapse card" id="{{ name }}">
    {% if step.is_pass %}
    <div class="card border-success">
    {% else %}
    <div class="card border-danger">
    {% endif %}
        {# Description #}
        {% if step.description %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.stepHeader.description }}</span></div>
        <div class="card-body">{{ markdown(step.description) }}</div>
        {% endif %}

        {% if not brief_mode %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.stepHeader.subStep }}</span></div>
        {% endif %}
        <ul class="list-group list-group-flush">
            {% for sub_step in step.sub_steps %}
            <li class="list-group-item px-{{ customize.padding.x }} py-{{ customize.padding.y }}">
                {{ render_sub_step(sub_step, '{}-sub-step-{}'.format(name, loop.index0), loop.index0) }}
            </li>
            {% endfor %}
        </ul>
    </div>
</div>
"""

_sub_step_tpl = """
{% if not brief_mode %}
{% if sub_step.is_pass %}
<a class="card-title btn d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    sub-step {{ index }}
    <span>{{ format_timedelta(sub_step.elapse) }}</span>
</a>
{% else %}
<a class="card-title text-white bg-danger btn btn-danger d-flex justify-content-between align-items-center" data-bs-toggle="collapse" href="#{{ name }}" role="button" aria-expanded="false" aria-controls="{{ name }}">
    sub-step {{ index }}
    <span>{{ format_timedelta(sub_step.elapse) }}</span>
</a>
{% endif %}
{% endif %}

{% if not brief_mode %}
<div class="collapse card" id="{{ name }}">
{% else %}
<div class="card" id="{{ name }}">
{% endif %}
    {% if sub_step.is_pass %}
    <div class="card border-success">
    {% else %}
    <div class="card border-danger">
    {% endif %}
        <div class="card-header"><span class="fw-bolder">{{ i18n.stepHeader.req }}</span></div>
        <div class="card-body">
            <div class="float-end">
                <button type="button" class="btn btn-sm py-0" onclick="copyToClipboard('{{ name }}-req')"
                    data-bs-toggle="tooltip" data-bs-placement="top" title="{{ i18n.toolTips.copy }}">
                    <i class="bi-clipboard"></i>
                </button>
            </div>
            <pre id="{{ name }}-req">{{ json.dumps(sub_step.req, indent=2) }}</pre>
        </div>

        {% if sub_step.is_pass or sub_step.is_err %}
        <div class="card-header justify-content-between d-flex">
        {% else %}
        <div class="card-header text-white bg-danger justify-content-between d-flex">
        {% endif %}
            <span class="fw-bolder">{{ i18n.stepHeader.res }}</span>
            <span>
                <span class="badge bg-success rounded-pill">{{ sub_step.assertion_pass }}</span>
                {% if sub_step.assertion_fail %}
                <span class="badge bg-danger rounded-pill">{{ sub_step.assertion_fail }}</span>
                {% endif %}
            </span>
        </div>
        <div class="card-body">
            <div class="float-end">
                <button type="button" class="btn btn-sm py-0" onclick="copyToClipboard('{{ name }}-res')"
                    data-bs-toggle="tooltip" data-bs-placement="top" title="{{ i18n.toolTips.copy }}">
                    <i class="bi-clipboard"></i>
                </button>
            </div>
            <pre id="{{ name }}-res">{{ format_sub_step_res(sub_step) }}</pre>
        </div>

        {% if sub_step.is_err %}
        <div class="card-header text-white bg-danger"><span class="fw-bolder">{{ i18n.stepHeader.err }}</span></div>
        <div class="card-body">
            <pre>{{ sub_step.err }}</pre>
        </div>
        {% endif %}
    </div>
</div>
"""


class HtmlReporter(Reporter):
    def __init__(self, args=None):
        super().__init__(args)
        args = merge(args, {
            "stepSeparator": "#",
            "font": {
                "style": """
<link rel="preconnect" href="https://fonts.googleapis.com">
<link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
<link href="https://fonts.googleapis.com/css2?family=JetBrains+Mono&family=Roboto+Condensed:wght@400;700&display=swap" rel="stylesheet">
""",
                "body": "'Roboto Condensed', sans-serif !important",
                "code": "'JetBrains Mono', monospace !important",
            },
            "extra": {
                "head": "",
                "bodyHeader": "",
                "bodyFooter": "",
            },
            "padding": {
                "x": 2,
                "y": 2,
            }
        })
        self.step_separator = args["stepSeparator"]
        self.customize = json.loads(json.dumps(args), object_hook=lambda x: SimpleNamespace(**x))

        env = Environment(loader=BaseLoader())
        env.globals.update(format_timedelta=HtmlReporter.format_timedelta)
        env.globals.update(json=json)
        env.globals.update(render_test=self.render_test)
        env.globals.update(render_case=self.render_case)
        env.globals.update(render_step=self.render_step)
        env.globals.update(render_sub_step=self.render_sub_step)
        env.globals.update(format_sub_step_res=self.format_sub_step_res)
        env.globals.update(brief_mode=True)
        env.globals.update(markdown=markdown.markdown)
        env.globals.update(i18n=self.i18n)
        env.globals.update(customize=self.customize)
        self.report_tpl = env.from_string(_report_tpl)
        self.test_tpl = env.from_string(_test_tpl)
        self.case_tpl = env.from_string(_case_tpl)
        self.step_tpl = env.from_string(_step_tpl)
        self.sub_step_tpl = env.from_string(_sub_step_tpl)

    def report(self, res: TestResult) -> str:
        return self.report_tpl.render(res=res)

    def render_test(self, res, name):
        return self.test_tpl.render(res=res, name=name)

    def render_case(self, case, name):
        return self.case_tpl.render(case=case, name=name)

    def render_step(self, step, name):
        return self.step_tpl.render(step=step, name=name)

    def render_sub_step(self, sub_step, name, index):
        return self.sub_step_tpl.render(sub_step=sub_step, name=name, index=index)

    def format_sub_step_res(self, sub_step: SubStepResult) -> str:
        return format_step_res(
            sub_step, separator=self.step_separator,
            pass_open="<span class='text-success'>", pass_close="</span>",
            fail_open="<span class='text-danger'>", fail_close="</span>",
        )

    @staticmethod
    def format_timedelta(t: datetime.timedelta):
        return "{:.3f}s".format(t.total_seconds())



#!/usr/bin/env python3


from datetime import datetime, timezone
from urllib.parse import urlparse
from dateutil import parser
import json
import uuid
import os


def to_time(val) -> datetime:
    return parser.parse(val)



#!/usr/bin/env python3


from elasticsearch import Elasticsearch
from datetime import datetime

from .driver import Driver
from ..util import merge, REQUIRED


class ElasticSearchDriver(Driver):
    def __init__(self, args):
        args = merge(args, {
            "endpoint": "http://localhost:9200",
            "username": "",
            "password": "",
        })

        if args["username"]:
            self.client = Elasticsearch(
                args["endpoint"].split(","),
                http_auth=(args["username"], args["password"])
            )
        else:
            self.client = Elasticsearch(
                args["endpoint"].split(","),
            )

    def name(self, req):
        return req["cmd"]

    def do(self, req):
        req = merge(req, {
            "cmd": REQUIRED,
        })

        do_map = {
            "index": self.index,
            "get": self.get,
            "search": self.search,
            "delete": self.delete,
        }

        if req["cmd"] not in do_map:
            raise Exception("unsupported cmd [{}]".format(req["cmd"]))

        return do_map[req["cmd"]](req)

    def index(self, req):
        req = merge(req, {
            "index": REQUIRED,
            "id": REQUIRED,
            "document": REQUIRED,
        })
        return self.client.index(index=req["index"], id=req["id"], document=req["document"])

    def get(self, req):
        req = merge(req, {
            "index": REQUIRED,
            "id": REQUIRED,
        })
        return self.client.get(index=req["index"], id=req["id"])

    def search(self, req):
        req = merge(req, {
            "index": REQUIRED,
            "query": {
                "match_all": {}
            }
        })

        return self.client.search(index=req["index"], query=req["query"])

    def delete(self, req):
        req = merge(req, {
            "index": REQUIRED,
            "id": REQUIRED,
        })
        return self.client.delete(index=req["index"], id=req["id"])



#!/usr/bin/env python3


from datetime import datetime
from aliyun.log import LogClient, PutLogsRequest, LogItem, GetLogsRequest

from ..util import merge, REQUIRED
from .driver import Driver


class SLSDriver(Driver):
    def __init__(self, args):
        args = merge(args, {
            "Endpoint": REQUIRED,
            "AccessKeyId": REQUIRED,
            "AccessKeySecret": REQUIRED,
        })

        self.client = LogClient(
            args["Endpoint"],
            args["AccessKeyId"],
            args["AccessKeySecret"],
        )

    def name(self, req):
        return req["Action"]

    def do(self, req: dict):
        req = merge(req, {
            "Action": REQUIRED
        })

        do_map = {
            "GetLogs": self.get_logs,
            "PutLogs": self.put_logs,
        }

        if req["Action"] not in do_map:
            raise Exception("unsupported action [{}]".format(req["Action"]))

        try:
            return do_map[req["Action"]](req)
        except Exception as e:
            raise e

    def get_logs(self, req):
        req = merge(req, {
            "Project": REQUIRED,
            "Logstore": REQUIRED,
            "FromTime": int(datetime.now().timestamp()) - 3600,
            "ToTime": int(datetime.now().timestamp()),
            "Topic": None,
            "Query": None,
            "Offset": 0,
            "Size": 100,
            "Reverse": False,
            "PowerSQL": False,
        })

        res = self.client.get_logs(GetLogsRequest(
            project=req["Project"],
            logstore=req["Logstore"],
            fromTime=req["FromTime"],
            toTime=req["ToTime"],
            topic=req["Topic"],
            query=req["Query"],
            line=req["Size"],
            offset=req["Offset"],
            reverse=req["Reverse"],
            power_sql=req["PowerSQL"],
        ))

        return {
            "RequestId": res.get_request_id(),
            "Header": res.headers,
            "Body": res.body,
        }

    def put_logs(self, req):
        req = merge(req, {
            "Project": REQUIRED,
            "Logstore": REQUIRED,
            "Topic": None,
            "Source": None,
            "HashKey": None,
            "Compress": False,
            "LogTags": {},
            "LogItems": [{
                "Timestamp": int(datetime.now().timestamp()),
                "Contents": {},
            }]
        })

        res = self.client.put_logs(PutLogsRequest(
            project=req["Project"],
            logstore=req["Logstore"],
            topic=req["Topic"],
            source=req["Source"],
            hashKey=req["HashKey"],
            compress=req["Compress"],
            logtags=req["LogTags"].items(),
            logitems=[
                LogItem(
                    timestamp=item["Timestamp"],
                    contents=[(k, v) for k, v in item["Contents"].items()],
                )
                for item in req["LogItems"]
            ]
        ))

        return {
            "RequestId": res.get_request_id(),
            "Header": res.headers,
            "Body": res.get_body(),
        }



#!/usr/bin/env python3

import json
import unittest

import tablestore

from .ots_driver import *


class TestOTSDriver(unittest.TestCase):
    def setUp(self) -> None:
        self.driver = OTSDriver(args={
            "Endpoint": "https://xx.cn-shanghai.ots.aliyuncs.com",
            "AccessKeyId": "xx",
            "AccessKeySecret": "xx",
            "Instance": "xx",
        })
        self.test_table_name = "TestQAS"

    def test_list_table(self):
        res = self.driver.do(req={
            "Action": "ListTable"
        })
        print(json.dumps(res))

    def test_crate_table(self):
        res = self.driver.do(req={
            "Action": "CreateTable",
            "TableMeta": {
                "TableName": self.test_table_name,
                "SchemaEntry": [{
                    "Name": "PK1",
                    "Type": "STRING",
                }, {
                    "Name": "PK2",
                    "Type": "STRING",
                }]
            },
            "TableOptions": {
                "TimeToLive": -1,
                "MaxVersion": 1,
                "MaxTimeDeviation": 86400,
            }
        })
        print(json.dumps(res))

    def test_delete_table(self):
        res = self.driver.do(req={
            "Action": "DeleteTable",
            "TableName": self.test_table_name,
        })
        print(json.dumps(res))

    def test_put_row(self):
        res = self.driver.do(req={
            "Action": "PutRow",
            "TableName": self.test_table_name,
            "Row": {
                "PrimaryKey": [{
                    "Key": "PK1",
                    "Val": "pkVal1",
                }, {
                    "Key": "PK2",
                    "Val": "pkVal2"
                }],
                "AttributeColumns": {
                    "key1": "val1",
                    "key2": "val2",
                }
            },
        })
        print(json.dumps(res))

    def test_get_row(self):
        res = self.driver.do(req={
            "Action": "GetRow",
            "TableName": self.test_table_name,
            "PrimaryKey": [{
                "Key": "PK1",
                "Val": "pkVal1",
            }, {
                "Key": "PK2",
                "Val": "pkVal2"
            }]
        })
        print(json.dumps(res))

    def test_delete_row(self):
        res = self.driver.do(req={
            "Action": "DeleteRow",
            "TableName": self.test_table_name,
            "PrimaryKey": [{
                "Key": "PK1",
                "Val": "pkVal1",
            }, {
                "Key": "PK2",
                "Val": "pkVal2"
            }]
        })
        print(json.dumps(res))

    def test_get_range(self):
        res = self.driver.do(req={
            "Action": "GetRange",
            "TableName": self.test_table_name,
            "StartPrimaryKey": [{
                "Key": "PK1",
            }, {
                "Key": "PK2",
            }],
            "EndPrimaryKey": [{
                "Key": "PK1",
            }, {
                "Key": "PK2",
            }]
        })
        print(json.dumps(res))

    def test_json(self):
        a = {
            "StartPrimaryKey": [{
                "Val": tablestore.INF_MIN
            }],
            "EndPrimaryKey": [{
                "Val": tablestore.INF_MAX,
            }]
        }
        print(tablestore.INF_MAX.__name__)
        print(json.dumps(a, default=lambda x: x.__name__))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3


import unittest
import smtplib
from email.mime.text import MIMEText
from email.header import Header

from .email_hook import EmailHook


class TestEmail(unittest.TestCase):
    def test_email(self):
        host = "smtp.qq.com"
        port = 465
        username = "hatlonely@foxmail.com"
        password = "******"
        receiver = "hatlonely@foxmail.com"
        subject = "Test 测试报告"
        body = "<p>hello world</p>"

        message = MIMEText(body, 'html', 'utf-8')
        message['From'] = Header(username, 'utf-8')
        message['To'] = Header(receiver, 'utf-8')
        message['Subject'] = Header(subject, 'utf-8')

        smtp = smtplib.SMTP_SSL(host, port)
        smtp.login(username, password)
        smtp.sendmail(username, receiver.split(";"), message.as_string())
        print("邮件发送成功")


class TestEmailHook(unittest.TestCase):
    def test_email(self):
        hook = EmailHook({
            "endpoint": "smtp.qq.com:465",
            "username": "hatlonely@foxmail.com",
            "password": "******",
            "receiver": "hatlonely@foxmail.com",
        })


def main():
    unittest.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3


import json
import durationpy
from colorama import Fore

from .hook import Hook
from ..result import TestResult, CaseResult
from ..reporter import format_step_res
from ..util import merge


class ReportHook(Hook):
    def __init__(self, args=None):
        super().__init__(args)
        args = merge(args, {
            "padding": "  "
        })
        self.padding_to_add = args["padding"]
        self.padding = ""

    def on_test_start(self, directory):
        print("{}{i18n.title.test} {directory}".format(self.padding, directory=directory, i18n=self.i18n))
        self.padding += self.padding_to_add

    def on_test_end(self, res: TestResult):
        self.padding = self.padding[:-len(self.padding_to_add)]

        if res.is_skip:
            print("{}{fore.YELLOW}{i18n.title.test} {res.name} {i18n.status.skip}{fore.RESET}".format(
                self.padding, res=res, i18n=self.i18n, fore=Fore,
            ))
            return

        if res.is_pass:
            print(
                "{}{fore.GREEN}{i18n.title.test} {res.name} {i18n.status.succ}{fore.RESET}，"
                "{i18n.summary.caseTotal}: {total_case}，"
                "{i18n.summary.casePass}: {fore.GREEN}{res.case_pass}{fore.RESET}，"
                "{i18n.summary.caseSkip}: {res.case_skip}，"
                "{i18n.summary.stepPass}: {fore.GREEN}{res.step_pass}{fore.RESET}，"
                "{i18n.summary.stepSkip}: {res.step_skip}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{res.assertion_pass}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    self.padding, total_case=res.case_pass + res.case_skip,
                    elapse=durationpy.to_str(res.elapse), res=res, i18n=self.i18n, fore=Fore,
                ),
            )
        else:
            if res.is_err:
                print('\n'.join([
                    "{}{}{}".format(self.padding_to_add, self.padding, line)
                    for line in "{i18n.testHeader.err} {res.err}".format(res=res, i18n=self.i18n).split("\n")
                ]))
            print(
                "{}{fore.RED}{i18n.title.test} {res.name} {i18n.status.fail}{fore.RESET}，"
                "{i18n.summary.caseTotal}: {total_case}，"
                "{i18n.summary.casePass}: {fore.GREEN}{res.case_pass}{fore.RESET}，"
                "{i18n.summary.caseSkip}: {res.case_skip}，"
                "{i18n.summary.caseFail}: {fore.RED}{res.case_fail}{fore.RESET}，"
                "{i18n.summary.stepPass}: {fore.GREEN}{res.step_pass}{fore.RESET}，"
                "{i18n.summary.stepSkip}: {res.step_skip}，"
                "{i18n.summary.stepFail}: {fore.RED}{res.step_fail}{fore.RESET}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{res.assertion_pass}{fore.RESET}，"
                "{i18n.summary.assertionFail}: {fore.RED}{res.assertion_fail}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    self.padding, total_case=res.case_pass + res.case_skip + res.case_fail,
                    elapse=durationpy.to_str(res.elapse), res=res, i18n=self.i18n, fore=Fore,
                ),
            )

    def on_case_end(self, res: CaseResult):
        print("\n".join([self.padding + i for i in self.format_case(res, "case")]))

    def on_set_up_end(self, res: CaseResult):
        print("\n".join([self.padding + i for i in self.format_case(res, "setUp")]))

    def on_tear_down_end(self, res: CaseResult):
        print("\n".join([self.padding + i for i in self.format_case(res, "tearDown")]))

    def format_case(self, res: CaseResult, case_type: str) -> list[str]:
        case_type_map = {
            "setUp": self.i18n.testHeader.setUp,
            "tearDown": self.i18n.testHeader.tearDown,
            "case": self.i18n.testHeader.case,
        }

        lines = []
        if res.is_skip:
            return ["{fore.YELLOW}{header} {res.name} {i18n.status.skip}{fore.RESET}".format(
                fore=Fore, res=res, i18n=self.i18n, header=case_type_map[case_type]
            )]

        if res.is_pass:
            lines.append(
                "{fore.GREEN}{header} {res.name} {i18n.status.succ}{fore.RESET}，"
                "{i18n.summary.stepPass}: {fore.GREEN}{res.step_pass}{fore.RESET}，"
                "{i18n.summary.stepSkip}: {res.step_skip}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{res.assertion_pass}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    header=case_type_map[case_type], elapse=durationpy.to_str(res.elapse), fore=Fore, i18n=self.i18n, res=res,
                ),
            )
        else:
            lines.append(
                "{fore.RED}{header} {res.name} {i18n.status.fail}{fore.RESET}，"
                "{i18n.summary.stepPass}: {fore.GREEN}{res.step_pass}{fore.RESET}，"
                "{i18n.summary.stepSkip}: {res.step_skip}，"
                "{i18n.summary.stepFail}: {fore.RED}{res.step_fail}{fore.RESET}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{res.assertion_pass}{fore.RESET}，"
                "{i18n.summary.assertionFail}: {fore.RED}{res.assertion_fail}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    header=case_type_map[case_type], elapse=durationpy.to_str(res.elapse), fore=Fore, i18n=self.i18n, res=res,
                ))

        for step in res.before_case_steps:
            lines.extend([self.padding_to_add + i for i in self.format_step(step, "beforeCaseStep")])
        for step in res.pre_steps:
            lines.extend([self.padding_to_add + i for i in self.format_step(step, "preStep")])
        for step in res.steps:
            lines.extend([self.padding_to_add + i for i in self.format_step(step, "step")])
        for step in res.post_steps:
            lines.extend([self.padding_to_add + i for i in self.format_step(step, "postStep")])
        for step in res.after_case_steps:
            lines.extend([self.padding_to_add + i for i in self.format_step(step, "afterCaseStep")])

        return lines

    def format_step(self, step, step_type: str) -> list[str]:
        step_type_map = {
            "beforeCaseStep": self.i18n.caseHeader.beforeCaseStep,
            "preStep": self.i18n.caseHeader.preStep,
            "step": self.i18n.caseHeader.step,
            "postStep": self.i18n.caseHeader.postStep,
            "afterCaseStep": self.i18n.caseHeader.afterCaseStep,
        }

        if step.is_skip:
            return ["{fore.YELLOW}{header} {step.name} {i18n.status.skip}{fore.RESET}".format(
                fore=Fore, step=step, i18n=self.i18n, header=step_type_map[step_type]
            )]

        lines = []
        if step.is_pass:
            lines.append(
                "{fore.GREEN}{header} {step.name} {i18n.status.succ}{fore.RESET}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{step.assertion_pass}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    fore=Fore, step=step, i18n=self.i18n, header=step_type_map[step_type], elapse=durationpy.to_str(step.elapse),
                ),
            )
        else:
            lines.append(
                "{fore.RED}{header} {step.name} {i18n.status.fail}{fore.RESET}，"
                "{i18n.summary.assertionPass}: {fore.GREEN}{step.assertion_pass}{fore.RESET}，"
                "{i18n.summary.assertionFail}: {fore.RED}{step.assertion_fail}{fore.RESET}，"
                "{i18n.summary.elapse}: {elapse}".format(
                    fore=Fore, step=step, i18n=self.i18n, header=step_type_map[step_type], elapse=durationpy.to_str(step.elapse),
                ),
            )

        for sub_step in step.sub_steps:
            lines.extend((self.i18n.stepHeader.req + ": " + json.dumps(sub_step.req, default=lambda x: str(x), indent=2)).split("\n"))
            if sub_step.is_err:
                lines.extend((self.i18n.stepHeader.res + ": " + json.dumps(sub_step.res, indent=2)).split("\n"))
                lines.extend((self.i18n.stepHeader.err + ": " + sub_step.err).split("\n"))
                return lines
            lines.extend((self.i18n.stepHeader.res + ": " + format_step_res(
                sub_step, pass_open=Fore.GREEN, pass_close=Fore.RESET, fail_open=Fore.RED, fail_close=Fore.RESET,
            )).split("\n"))
        return lines




#!/usr/bin/env python3


from .include import *
import subprocess


def exec_with_res(rule, val=None, case=None, step=None, var=None, x=None):
    loc = {}
    env = globals()
    env.update(val=val)
    env.update(case=case)
    env.update(step=step)
    env.update(var=var)
    env.update(x=x)
    env.update(to_time=to_time)
    exec(rule, env, loc)
    return loc["res"]


def exec_shell(rule):
    process = subprocess.run(
        ["/bin/bash", "-c", rule],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    return process.stdout.decode("utf-8")



#!/usr/bin/env python3


import copy
import itertools
import re
import time
import yaml
import traceback
import os
import json
import importlib
import sys
import pathlib
from types import SimpleNamespace
from datetime import datetime
from dataclasses import dataclass
import concurrent.futures
from itertools import repeat
from pathlib import Path


from ..driver import driver_map, Driver
from ..reporter import reporter_map
from ..hook import hook_map, Hook
from ..assertion import expect, check
from ..util import render, merge, REQUIRED
from ..result import TestResult, CaseResult, StepResult, SubStepResult
from .retry_until import Retry, Until, RetryError, UntilError
from .generate import generate_req, generate_res, calculate_num, grouper


@dataclass
class RuntimeConstant:
    test_directory: str
    case_directory: str
    case_regex: str
    case_name: str
    case_id: str
    skip_set_up: bool
    skip_tear_down: bool
    parallel: bool
    x: str


@dataclass
class RuntimeContext:
    ctx: dict[str, Driver]
    var: SimpleNamespace
    var_info: dict
    dft: dict
    common_step_info: dict
    before_case_info: list
    after_case_info: list
    driver_map: dict
    x: any
    hooks: list[Hook]
    step_pool: concurrent.futures.ThreadPoolExecutor
    case_pool: concurrent.futures.ThreadPoolExecutor
    test_pool: concurrent.futures.ThreadPoolExecutor


class Framework:
    constant: RuntimeConstant

    def __init__(
        self,
        test_directory=None,
        case_directory=None,
        case_name=None,
        case_id=None,
        case_regex=None,
        skip_set_up=False,
        skip_tear_down=False,
        reporter="text",
        x=None,
        json_result=None,
        parallel=False,
        step_pool_size=None,
        case_pool_size=None,
        test_pool_size=None,
        hook=None,
        customize=None,
        lang=None,
    ):
        self.step_pool = None
        self.case_pool = None
        self.test_pool = None
        if parallel:
            self.step_pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=step_pool_size) if step_pool_size else concurrent.futures.ThreadPoolExecutor()
            self.case_pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=case_pool_size) if case_pool_size else concurrent.futures.ThreadPoolExecutor()
            self.test_pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=test_pool_size) if test_pool_size else concurrent.futures.ThreadPoolExecutor()

        test_directory = test_directory.strip().rstrip("/") if test_directory else test_directory
        self.constant = RuntimeConstant(
            test_directory=test_directory,
            case_directory=test_directory if not case_directory else os.path.join(test_directory, case_directory.strip().rstrip("/")),
            case_regex=case_regex,
            case_name=case_name,
            case_id=case_id,
            skip_set_up=skip_set_up,
            skip_tear_down=skip_tear_down,
            parallel=parallel,
            x=x,
        )
        self.reporter_map = reporter_map
        self.driver_map = driver_map
        self.hook_map = hook_map
        self.x = None
        if x:
            self.x = Framework.load_x(x)
            if hasattr(self.x, "reporter_map"):
                self.reporter_map = self.reporter_map | self.x.reporter_map
            if hasattr(self.x, "driver_map"):
                self.driver_map = self.driver_map | self.x.driver_map
            if hasattr(self.x, "hook_map"):
                self.hook_map = self.hook_map | self.x.hook_map

        if not customize and os.path.exists("{}/.qas/customize.yaml".format(Path.home())):
            customize = "{}/.qas/customize.yaml".format(Path.home())

        hooks = hook.split(",") if hook else []
        cfg = {}
        if customize:
            with open(customize, "r", encoding="utf-8") as fp:
                cfg = yaml.safe_load(fp)
        cfg = merge(cfg, {
            "i18n": {},
            "reporter": {
                reporter: {
                    "i18n": {}
                },
            },
            "hook": dict([(i, {
                "i18n": {}
            }) for i in hooks]),
            "framework": {
                "keyPrefix": {
                    "eval": "#",
                    "exec": "%",
                    "loop": "!",
                    "shell": "$",
                },
                "loadingFiles": {
                    "ctx": "ctx.yaml",
                    "var": "var.yaml",
                    "setUp": "set_up.yaml",
                    "tearDown": "tear_down.yaml",
                    "beforeCase": "before_case.yaml",
                    "afterCase": "after_case.yaml",
                    "commonStep": "common_step.yaml",
                    "description": "README.md",
                }
            }
        })
        cfg = render(
            cfg,
            peval=cfg["framework"]["keyPrefix"]["eval"],
            pexec=cfg["framework"]["keyPrefix"]["exec"],
            pshell=cfg["framework"]["keyPrefix"]["shell"],
        )

        if "i18n" in cfg:
            cfg["reporter"][reporter] = merge(cfg["reporter"][reporter], cfg["i18n"])
            for key in hooks:
                cfg["hook"][key] = merge(cfg["hook"][key], cfg["i18n"])
        if lang:
            cfg["reporter"][reporter]["lang"] = lang
            for key in hooks:
                cfg["hook"][key]["lang"] = lang

        self.customize = json.loads(json.dumps(cfg["framework"]), object_hook=lambda y: SimpleNamespace(**y))

        self.reporter = self.reporter_map[reporter](cfg["reporter"][reporter])
        self.hooks = [self.hook_map[i](cfg["hook"][i]) for i in hooks]

        self.json_result = json_result

    def format(self):
        res = TestResult.from_json(json.load(open(self.json_result)))
        print(self.reporter.report(res))

    def run(self):
        rctx = RuntimeContext(
            ctx={},
            dft={},
            var_info={},
            var=None,
            common_step_info={},
            before_case_info=[],
            after_case_info=[],
            driver_map=self.driver_map,
            x=self.x,
            hooks=self.hooks,
            step_pool=self.step_pool,
            case_pool=self.case_pool,
            test_pool=self.test_pool,
        )

        res = self.must_run_test(self.constant.test_directory, self.customize, self.constant, rctx)
        print(self.reporter.report(res))

        for hook in self.hooks:
            hook.on_exit(res)

        return res.is_pass

    @staticmethod
    def must_run_test(
        directory: str,
        customize,
        constant: RuntimeConstant,
        rctx: RuntimeContext,
    ):
        for hook in rctx.hooks:
            hook.on_test_start(directory)
        try:
            result = Framework.run_test(directory, customize, constant, rctx)
        except Exception as e:
            result = TestResult(directory, directory, "", "Exception {}".format(traceback.format_exc()))
        for hook in rctx.hooks:
            hook.on_test_end(result)
        return result

    @staticmethod
    def run_test(
        directory: str,
        customize,
        constant: RuntimeConstant,
        parent_rctx: RuntimeContext,
    ):
        if constant.case_id and not (constant.case_directory + "/").startswith(directory + "/"):
            return TestResult(directory, directory[len(constant.test_directory)+1:], "", is_skip=True)
        if not (constant.case_directory + "/").startswith(directory + "/") and \
                not (directory + "/").startswith(constant.case_directory + "/"):
            return TestResult(directory, directory[len(constant.test_directory)+1:], "", is_skip=True)

        now = datetime.now()

        info = Framework.load_ctx(os.path.basename(directory), "{}/{}".format(directory, customize.loadingFiles.ctx))
        description = info["description"] + Framework.load_description("{}/{}".format(directory, customize.loadingFiles.description))
        var_info = copy.deepcopy(parent_rctx.var_info) | info["var"] | Framework.load_var("{}/{}".format(directory, customize.loadingFiles.var))
        var_info = render(var_info, peval=customize.keyPrefix.eval, pexec=customize.keyPrefix.exec, pshell=customize.keyPrefix.shell)
        var = json.loads(json.dumps(var_info), object_hook=lambda x: SimpleNamespace(**x))
        common_step_info = copy.deepcopy(parent_rctx.common_step_info) | info["commonStep"] | Framework.load_common_step("{}/{}".format(directory, customize.loadingFiles.commonStep))
        before_case_info = copy.deepcopy(parent_rctx.before_case_info) + info["beforeCase"] + list(Framework.load_step("{}/{}".format(directory, customize.loadingFiles.beforeCase)))
        after_case_info = info["afterCase"] + list(Framework.load_step("{}/{}".format(directory, customize.loadingFiles.afterCase))) + copy.deepcopy(parent_rctx.after_case_info)
        ctx = copy.copy(parent_rctx.ctx)
        dft = copy.deepcopy(parent_rctx.dft)
        for key in info["ctx"]:
            val = merge(info["ctx"][key], {
                "type": REQUIRED,
                "args": {},
                "dft": {
                    "req": {},
                    "retry": {
                        "attempts": 1,
                        "delay": "1s",
                    },
                    "until": {
                        "cond": "",
                        "attempts": 5,
                        "delay": "1s",
                    },
                },
            })
            val = render(val, var=var, x=parent_rctx.x, peval=customize.keyPrefix.eval, pexec=customize.keyPrefix.exec, pshell=customize.keyPrefix.shell)
            ctx[key] = parent_rctx.driver_map[val["type"]](val["args"])
            dft[key] = val["dft"]

        test_result = TestResult(directory, info["name"], description)

        rctx = RuntimeContext(
            ctx=ctx,
            var=var,
            var_info=var_info,
            dft=dft,
            common_step_info=common_step_info,
            before_case_info=before_case_info,
            after_case_info=after_case_info,
            driver_map=driver_map,
            x=parent_rctx.x,
            hooks=parent_rctx.hooks,
            step_pool=parent_rctx.step_pool,
            case_pool=parent_rctx.case_pool,
            test_pool=parent_rctx.test_pool,
        )

        # 执行 set_up
        if not constant.skip_set_up:
            if not constant.parallel:
                for case_info in Framework.set_ups(customize, info, directory):
                    result = Framework.must_run_case(directory, customize, constant, rctx, case_info, case_type="set_up")
                    test_result.add_set_up_result(result)
            else:
                # 并发执行，每次执行 ctx.yaml 中 parallel.setUp 定义的个数
                for i in grouper(Framework.set_ups(customize, info, directory), info["parallel"]["setUp"]):
                    results = parent_rctx.case_pool.map(
                        Framework.must_run_case,
                        repeat(directory), repeat(customize), repeat(constant), repeat(rctx), i, repeat("set_up")
                    )
                    for result in results:
                        test_result.add_set_up_result(result)
            if not test_result.is_pass:
                return test_result

        # 执行 case
        if directory.startswith(constant.case_directory):
            if not constant.parallel:
                for case_info in Framework.cases(customize, constant, info, directory):
                    result = Framework.must_run_case(directory, customize, constant, rctx, case_info)
                    test_result.add_case_result(result)
            else:
                # 并发执行，每次执行 ctx.yaml 中 parallel.case 定义的个数
                for i in grouper(Framework.cases(customize, constant, info, directory), info["parallel"]["case"]):
                    results = parent_rctx.case_pool.map(
                        Framework.must_run_case,
                        repeat(directory), repeat(customize), repeat(constant), repeat(rctx), i
                    )
                    for result in results:
                        test_result.add_case_result(result)

        # 执行子目录
        if not constant.parallel:
            for sub_directory in [os.path.join(directory, i) for i in os.listdir(directory) if os.path.isdir(os.path.join(directory, i))]:
                result = Framework.must_run_test(sub_directory, customize, constant, rctx)
                test_result.add_sub_test_result(result)
        else:
            # 并发执行，每次执行 ctx.yaml 中 parallel.subTest 定义的个数
            # 每次执行会递归地使用 test_pool，每层目录需要占用一个线程，当 pool size 小于目录嵌套层数时会导致死锁
            # 不设置 test-pool-size 或者设置 test-pool-size 大于最大嵌套层数可解决这个问题
            for directories in grouper([
                os.path.join(directory, i)
                for i in os.listdir(directory)
                if os.path.isdir(os.path.join(directory, i))
            ], info["parallel"]["subTest"]):
                results = parent_rctx.test_pool.map(
                    Framework.must_run_test, directories, repeat(customize), repeat(constant), repeat(rctx),
                )
                for result in results:
                    test_result.add_sub_test_result(result)

        # 执行 tear_down
        if not constant.skip_tear_down:
            if not constant.parallel:
                for case_info in Framework.tear_downs(customize, info, directory):
                    result = Framework.must_run_case(directory, customize, constant, rctx, case_info, case_type="tear_down")
                    test_result.add_tear_down_result(result)
            else:
                # 并发执行，每次执行 ctx.yaml 中 parallel.tearDown 定义的个数
                for i in grouper(Framework.tear_downs(customize, info, directory), info["parallel"]["tearDown"]):
                    results = parent_rctx.case_pool.map(
                        Framework.must_run_case,
                        repeat(directory), repeat(customize), repeat(constant), repeat(rctx), i, repeat("tear_down")
                    )
                    for result in results:
                        test_result.add_tear_down_result(result)

        test_result.elapse = datetime.now() - now
        return test_result

    @staticmethod
    def need_skip(constant, case, var, case_type):
        if case_type == "set_up" or case_type == "tear_down":
            return False
        if constant.case_name and constant.case_name != case["name"]:
            return True
        if constant.case_regex and not re.search(constant.case_regex, case["name"]):
            return True
        if "cond" in case and case["cond"] and not check(case["cond"], var=var):
            return True
        return False

    @staticmethod
    def set_ups(customize, info, directory):
        for case in info["setUp"]:
            yield case
        if os.path.isfile("{}/{}".format(directory, customize.loadingFiles.setUp)):
            for case in Framework.load_case(directory, customize.loadingFiles.setUp):
                yield case

    @staticmethod
    def tear_downs(customize, info, directory):
        for case in info["tearDown"]:
            yield case
        if os.path.isfile("{}/{}".format(directory, customize.loadingFiles.tearDown)):
            for case in Framework.load_case(directory, customize.loadingFiles.tearDown):
                yield case

    @staticmethod
    def cases(customize, constant: RuntimeConstant, info, directory):
        if constant.case_id:
            pos = constant.case_id.find("-")
            if pos == -1:
                filename = constant.case_id + ".yaml"
                idx = 0
            else:
                filename = constant.case_id[:pos] + ".yaml"
                idx = int(constant.case_id[pos+1:])
            if filename == customize.loadingFiles.ctx:
                yield Framework.format_case(info["case"][idx], filename, idx)
            else:
                yield Framework.format_case(list(Framework.load_case(directory, filename))[idx], filename, idx)
            return

        for idx, case in enumerate(info["case"]):
            yield Framework.format_case(case, customize.loadingFiles.ctx, idx)

        for filename in [
            i
            for i in os.listdir(directory)
            if i not in [
                customize.loadingFiles.ctx,
                customize.loadingFiles.var,
                customize.loadingFiles.setUp,
                customize.loadingFiles.tearDown,
                customize.loadingFiles.beforeCase,
                customize.loadingFiles.afterCase,
                customize.loadingFiles.commonStep,
            ]
            and os.path.isfile(os.path.join(directory, i))
        ]:
            if not filename.endswith(".yaml"):
                continue
            for case in Framework.load_case(directory, filename):
                yield case

    @staticmethod
    def load_x(filename):
        if not os.path.exists(filename) or not os.path.isdir(filename):
            return {}
        p = pathlib.Path(filename)
        sys.path.append(str(p.parent.absolute()))
        return importlib.import_module(str(p.name), "x")

    @staticmethod
    def load_var(filename):
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return {}
        with open(filename, "r", encoding="utf-8") as fp:
            info = yaml.safe_load(fp)
        return info

    @staticmethod
    def load_description(filename):
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return ""
        with open(filename, "r", encoding="utf-8") as fp:
            info = fp.readlines()
        return info

    @staticmethod
    def load_ctx(name, filename):
        dft = {
            "name": name,
            "description": "",
            "parallel": {
                "case": 0,
                "subTest": 0,
                "setUp": 0,
                "tearDown": 0,
            },
            "ctx": {},
            "var": {},
            "case": [],
            "setUp": [],
            "tearDown": [],
            "beforeCase": [],
            "afterCase": [],
            "commonStep": {},
        }
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return dft
        with open(filename, "r", encoding="utf-8") as fp:
            info = yaml.safe_load(fp)
        return merge(info, dft)

    @staticmethod
    def load_case(directory, filename):
        with open(os.path.join(directory, filename), "r", encoding="utf-8") as fp:
            info = yaml.safe_load(fp)
            if isinstance(info, dict):
                yield Framework.format_case(info, filename, 0)
            if isinstance(info, list):
                for idx, item in enumerate(info):
                    yield Framework.format_case(item, filename, idx)

    @staticmethod
    def format_case(info, filename, idx):
        text = os.path.splitext(filename)[0]
        case_id = "{}-{}".format(text, idx)
        if idx == 0 and text.find("-") == -1:
            case_id = text
        info = merge(info, {
            "name": REQUIRED,
            "description": "",
            "cond": "",
            "caseID": case_id
        })
        return info

    @staticmethod
    def load_step(filename):
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return []
        with open(filename, "r", encoding="utf-8") as fp:
            info = yaml.safe_load(fp)
            if not info:
                return []
            for step in info:
                yield step

    @staticmethod
    def load_common_step(filename):
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return {}
        with open(filename, "r", encoding="utf-8") as fp:
            info = yaml.safe_load(fp)
            if not info:
                return {}
        return info

    @staticmethod
    def must_run_case(directory, customize, constant: RuntimeConstant, rctx: RuntimeContext, case_info, case_type="case"):
        case_info = merge(case_info, {
            "name": REQUIRED,
            "description": "",
            "cond": "",
            "label": {},
            "preStep": [],
            "postStep": [],
        })

        for hook in rctx.hooks:
            if case_type == "set_up":
                hook.on_set_up_start(case_info)
            elif case_type == "tear_down":
                hook.on_tear_down_start(case_info)
            else:
                hook.on_case_start(case_info)
        result = Framework.run_case(directory, customize, constant, rctx, case_info, case_type=case_type)
        for hook in rctx.hooks:
            if case_type == "set_up":
                hook.on_set_up_end(result)
            elif case_type == "tear_down":
                hook.on_tear_down_end(result)
            else:
                hook.on_case_end(result)
        return result

    @staticmethod
    def run_case(directory, customize, constant: RuntimeConstant, rctx: RuntimeContext, case_info, case_type="case"):
        if Framework.need_skip(constant, case_info, rctx.var, case_type):
            return CaseResult(directory=directory, id_=case_info["caseID"], name=case_info["name"], is_skip=True)

        command = ""
        case_id = ""
        if case_type == "case":
            case_id = case_info["caseID"]
            if constant.x:
                command = 'qas -x "{}" -t "{}" -c "{}" --case-id "{}"'.format(
                    constant.x, constant.test_directory, directory[len(constant.test_directory) + 1:], case_id,
                )
            else:
                command = 'qas -t "{}" -c "{}" --case-id "{}"'.format(
                    constant.test_directory, directory[len(constant.test_directory) + 1:], case_id,
                )
        case = CaseResult(directory=directory, id_=case_id, name=case_info["name"], description=case_info["description"], command=command)

        now = datetime.now()

        if case_type == "case":
            for idx, step_info, case_add_step_func in itertools.chain([list(i) + [case.add_before_case_step_result] for i in enumerate(rctx.before_case_info)]):
                step = Framework.must_run_step(customize, constant, rctx, step_info, case)
                case_add_step_func(step)
                if not step.is_pass:
                    break
            if not case.is_pass:
                return case

        for idx, step_info, case_add_step_func in itertools.chain(
            [list(i) + [case.add_case_pre_step_result] for i in enumerate([rctx.common_step_info[i] for i in case_info["preStep"]])],
            [list(i) + [case.add_case_step_result] for i in enumerate(case_info["step"])],
            [list(i) + [case.add_case_post_step_result] for i in enumerate([rctx.common_step_info[i] for i in case_info["postStep"]])],
        ):
            step = Framework.must_run_step(customize, constant, rctx, step_info, case)
            case_add_step_func(step)
            if not step.is_pass:
                break

        if case_type == "case":
            for idx, step_info, case_add_step_func in itertools.chain([list(i) + [case.add_after_case_step_result] for i in enumerate(rctx.after_case_info)]):
                step = Framework.must_run_step(customize, constant, rctx, step_info, case)
                case_add_step_func(step)
                if not step.is_pass:
                    break

        case.elapse = datetime.now() - now
        return case

    @staticmethod
    def must_run_step(customize, constant: RuntimeConstant, rctx: RuntimeContext, step_info, case):
        step_info = merge(step_info, {
            "name": "",
            "description": "",
            "parallel": 0,
            "res": {},
            "retry": {},
            "until": {},
            "cond": "",
        })

        for hook in rctx.hooks:
            hook.on_step_start(step_info)
        step = Framework.run_step(customize, constant, rctx, step_info, case)
        for hook in rctx.hooks:
            hook.on_step_end(step)
        return step

    @staticmethod
    def run_step(customize, constant, rctx, step_info, case):
        # 条件步骤
        if step_info["cond"] and not check(step_info["cond"], case=case, var=rctx.var, x=rctx.x):
            return StepResult(step_info["name"], step_info["ctx"], is_skip=True)
        step = StepResult(step_info["name"], step_info["ctx"], step_info["description"])
        now = datetime.now()

        # use json transform tuple to list
        step_info["req"] = render(json.loads(json.dumps(step_info["req"])), case=case, var=rctx.var, x=rctx.x, peval=customize.keyPrefix.eval, pexec=customize.keyPrefix.exec, pshell=customize.keyPrefix.shell)

        mode = ""
        if customize.keyPrefix.eval + "res" in step_info:
            step_info["res"] = step_info[customize.keyPrefix.eval + "res"]
            mode = customize.keyPrefix.eval
        elif customize.keyPrefix.exec + "res" in step_info:
            step_info["res"] = step_info[customize.keyPrefix.exec + "res"]
            mode = customize.keyPrefix.exec
        if not constant.parallel:
            for req, res in zip(
                generate_req(step_info["req"], p=customize.keyPrefix.loop),
                generate_res(step_info["res"], calculate_num(step_info["req"], p=customize.keyPrefix.loop), p=customize.keyPrefix.loop)
            ):
                result = Framework.run_sub_step(customize, rctx, case, req, res, step_info, mode=mode)
                step.add_sub_step_result(result)
        else:
            # 并发执行，每次执行 case.step 中 parallel 定义的个数
            for reqs, ress in zip(
                    grouper(generate_req(step_info["req"], p=customize.keyPrefix.loop), step_info["parallel"]),
                    grouper(generate_res(step_info["res"], calculate_num(step_info["req"], p=customize.keyPrefix.loop), p=customize.keyPrefix.loop), step_info["parallel"])
            ):
                results = rctx.step_pool.map(
                    Framework.run_sub_step,
                    repeat(customize), repeat(rctx), repeat(case), reqs, ress, repeat(step_info), repeat(mode)
                )
                for result in results:
                    step.add_sub_step_result(result)

        # auto name step
        if not step.name and step.req:
            step.name = rctx.ctx[step_info["ctx"]].name(step.req)
        if not step.name:
            step.name = "anonymous-step"
        step.elapse = datetime.now() - now
        return step

    @staticmethod
    def run_sub_step(customize, rctx: RuntimeContext, case, req, res, step_info, mode=""):
        sub_step_result = SubStepResult()
        sub_step_start = datetime.now()
        try:
            req = merge(req, rctx.dft[step_info["ctx"]]["req"])
            sub_step_result.req = req

            retry = Retry(merge(step_info["retry"], rctx.dft[step_info["ctx"]]["retry"]))
            until = Until(merge(step_info["until"], rctx.dft[step_info["ctx"]]["until"]))

            for i in range(until.attempts):
                for j in range(retry.attempts):
                    step_res = rctx.ctx[step_info["ctx"]].do(req)
                    sub_step_result.res = step_res
                    if retry.condition == "" or not check(retry.condition, case=case, step=sub_step_result, var=rctx.var, x=rctx.x):
                        break
                    time.sleep(retry.delay.total_seconds())
                else:
                    raise RetryError()
                if until.condition == "" or check(until.condition, case=case, step=sub_step_result, var=rctx.var, x=rctx.x):
                    break
                time.sleep(until.delay.total_seconds())
            else:
                raise UntilError()

            result = expect(step_res, json.loads(json.dumps(res)), case=case, step=sub_step_result, var=rctx.var, x=rctx.x, peval=customize.keyPrefix.eval, pexec=customize.keyPrefix.exec, mode=mode)
            sub_step_result.add_expect_result(result)
        except RetryError as e:
            sub_step_result.set_error("RetryError [{}]".format(retry))
        except UntilError as e:
            sub_step_result.set_error("UntilError [{}], ".format(until))
        except Exception as e:
            sub_step_result.set_error("Exception {}".format(traceback.format_exc()))
        # ensure req can json serialize
        sub_step_result.req = json.loads(json.dumps(sub_step_result.req, default=lambda y: str(y)))
        sub_step_result.elapse = datetime.now() - sub_step_start
        return sub_step_result



#!/usr/bin/env python3


from datetime import datetime
from aliyun.log import LogClient, PutLogsRequest, LogItem, GetLogsRequest

from .hook import Hook
from ..result import TestResult, CaseResult, StepResult
from ..util import merge, REQUIRED


class SLSLogHook(Hook):
    def __init__(self, args):
        super().__init__(args)

        args = merge(args, {
            "Endpoint": REQUIRED,
            "AccessKeyId": REQUIRED,
            "AccessKeySecret": REQUIRED,
            "Project": REQUIRED,
            "Logstore": REQUIRED,
            "Topic": None,
            "Source": None,
            "HashKey": None,
            "LogTags": {}
        })

        self.client = LogClient(
            args["Endpoint"],
            args["AccessKeyId"],
            args["AccessKeySecret"],
        )
        self.project = args["Project"]
        self.logstore = args["Logstore"]
        self.topic = args["Topic"]
        self.source = args["Source"]
        self.hash_key = args["HashKey"]
        self.log_tags = args["LogTags"]

    def on_exit(self, res: TestResult):
        self.log(res.name, "test", res.status, res.elapse, {
            "casePass": res.case_pass,
            "caseFail": res.case_fail,
            "caseSkip": res.case_skip,
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_test_end(self, res: TestResult):
        self.log(res.name, "subTest", res.status, res.elapse, {
            "casePass": res.case_pass,
            "caseFail": res.case_fail,
            "caseSkip": res.case_skip,
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_set_up_end(self, res: CaseResult):
        self.log(res.name, "setUp", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_case_end(self, res: CaseResult):
        self.log(res.name, "case", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_tear_down_end(self, res: CaseResult):
        self.log(res.name, "tearDown", res.status, res.elapse, {
            "stepPass": res.step_pass,
            "stepFail": res.step_fail,
            "stepSkip": res.step_skip,
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def on_step_end(self, res: StepResult):
        if res.is_skip:
            status = "skip"
        elif res.is_pass:
            status = "pass"
        else:
            status = "fail"
        self.log(res.name, "step", status, res.elapse, {
            "assertionPass": res.assertion_pass,
            "assertionFail": res.assertion_fail,
        })

    def log(self, name, type_, status, elapse, detail):
        self.client.put_logs(PutLogsRequest(
            project=self.project,
            logstore=self.logstore,
            topic=self.topic,
            source=self.source,
            hashKey=self.hash_key,
            logtags=self.log_tags.items(),
            logitems=[
                LogItem(
                    timestamp=int(datetime.now().timestamp()),
                    contents=[
                        ("type", type_),
                        ("status", status),
                        ("elapseMs", str(int(elapse.total_seconds() * 1000))),
                        *[(k, str(v)) for k, v in detail.items()],
                    ],
                )
            ]
        ))



#!/usr/bin/env python3


import datetime
import durationpy
from ..util import merge


class RetryError(Exception):
    pass


class UntilError(Exception):
    pass


class Retry:
    attempts: int
    delay: datetime.timedelta
    condition: str

    def __init__(self, args):
        args = merge(args, {
            "attempts": 1,
            "delay": "1s",
            "cond": "",
        })
        self.attempts = args["attempts"]
        self.delay = durationpy.from_str(args["delay"])
        self.condition = args["cond"]

    def __repr__(self):
        return "cond: [{}], attempts: {}, delay: {}".format(self.condition, self.attempts, durationpy.to_str(self.delay))


class Until:
    attempts: int
    delay: datetime.timedelta
    condition: str

    def __init__(self, args):
        args = merge(args, {
            "attempts": 1,
            "delay": "1s",
            "cond": "",
        })
        self.attempts = args["attempts"]
        self.delay = durationpy.from_str(args["delay"])
        self.condition = args["cond"]

    def __repr__(self):
        return "cond: [{}], attempts: {}, delay: {}".format(self.condition, self.attempts, durationpy.to_str(self.delay))



#!/usr/bin/env python3


import tablestore
import json
import traceback

from ..util import merge, REQUIRED
from .driver import Driver


class OTSDriver(Driver):
    client: tablestore.OTSClient = None

    def __init__(self, args: dict):
        args = merge(args, {
            "Endpoint": REQUIRED,
            "AccessKeyId": REQUIRED,
            "AccessKeySecret": REQUIRED,
            "Instance": REQUIRED,
        })

        self.client = tablestore.OTSClient(args["Endpoint"], args["AccessKeyId"], args["AccessKeySecret"], args["Instance"])

    def name(self, req):
        return req["Action"]

    def do(self, req: dict):
        req = merge(req, {
            "Action": REQUIRED
        })

        do_map = {
            "CreateTable": self.create_table,
            "DeleteTable": self.delete_table,
            "ListTable": self.list_table,
            "GetRow": self.get_row,
            "PutRow": self.put_row,
            "DeleteRow": self.delete_row,
            "GetRange": self.get_range,
        }
        if req["Action"] not in do_map:
            raise Exception("unsupported action [{}]".format(req["Action"]))
        try:
            return do_map[req["Action"]](req)
        except tablestore.error.OTSServiceError as e:
            return {
                "Code": e.error_code,
                "Message": e.message,
                "RequestId": e.request_id,
                "Status": e.http_status
            }
        except Exception as e:
            return {
                "Exception": traceback.format_exc(),
            }

    def list_table(self, req):
        res = self.client.list_table()
        return json.loads(json.dumps(res))

    def create_table(self, req):
        req = merge(req, {
            "TableMeta": {
                "TableName": REQUIRED,
                "SchemaEntry": [{
                    "Name": REQUIRED,
                    "Type": REQUIRED,  # STRING / INTEGER / BOOLEAN / DOUBLE / BINARY
                }],
                "DefinedColumns": []
            },
            "TableOptions": {
                "TimeToLive": -1,
                "MaxVersion": 1,
                "MaxTimeDeviation": 86400,
            }
        })

        self.client.create_table(
            table_meta=tablestore.TableMeta(
                table_name=req["TableMeta"]["TableName"],
                schema_of_primary_key=[(i["Name"], i["Type"]) for i in req["TableMeta"]["SchemaEntry"]],
                defined_columns=[(i["Name"], i["Type"]) for i in req["TableMeta"]["DefinedColumns"]],
            ),
            table_options=tablestore.TableOptions(
                time_to_live=req["TableOptions"]["TimeToLive"],
                max_version=req["TableOptions"]["MaxVersion"],
                max_time_deviation=req["TableOptions"]["MaxTimeDeviation"],
            ),
            reserved_throughput=tablestore.ReservedThroughput(tablestore.CapacityUnit(0, 0))
        )
        return {}

    def delete_table(self, req):
        req = merge(req, {
            "TableName": REQUIRED,
        })
        self.client.delete_table(req["TableName"])
        return {}

    def put_row(self, req):
        req = merge(req, {
            "TableName": REQUIRED,
            "Row": {
                "PrimaryKey": [{
                    "Key": REQUIRED,
                    "Val": REQUIRED,
                }]
            },
            "Condition": "IGNORE",  # IGNORE / EXPECT_EXIST / EXPECT_NOT_EXIST
        })

        _, _ = self.client.put_row(
            table_name=req["TableName"],
            row=tablestore.Row(
                primary_key=[(i["Key"], i["Val"]) for i in req["Row"]["PrimaryKey"]],
                attribute_columns=req["Row"]["AttributeColumns"].items(),
            ),
            condition=tablestore.Condition(req["Condition"])
        )

        return {}

    def get_row(self, req):
        req = merge(req, {
            "TableName": REQUIRED,
            "PrimaryKey": [{
                "Key": REQUIRED,
                "Val": REQUIRED,
            }],
            "MaxVersion": 1,
        })

        _, row, _ = self.client.get_row(
            table_name=req["TableName"],
            primary_key=[(i["Key"], i["Val"]) for i in req["PrimaryKey"]],
            max_version=req["MaxVersion"],
        )

        return dict([(i[0], i[1]) for i in row.attribute_columns])

    def delete_row(self, req):
        req = merge(req, {
            "TableName": REQUIRED,
            "PrimaryKey": [{
                "Key": REQUIRED,
                "Val": REQUIRED,
            }],
            "Condition": "IGNORE",  # IGNORE / EXPECT_EXIST / EXPECT_NOT_EXIST
        })

        self.client.delete_row(
            table_name=req["TableName"],
            row=tablestore.Row(
                primary_key=[(i["Key"], i["Val"]) for i in req["PrimaryKey"]],
            ),
            condition=tablestore.Condition(req["Condition"])
        )

        return {}

    def get_range(self, req):
        req = merge(req, {
            "TableName": REQUIRED,
            "Direction": "FORWARD",
            "StartPrimaryKey": [{
                "Key": REQUIRED,
                "Val": tablestore.INF_MIN
            }],
            "EndPrimaryKey": [{
                "Key": REQUIRED,
                "Val": tablestore.INF_MAX,
            }],
            "MaxVersion": 1,
        })

        _, _, rows, _ = self.client.get_range(
            table_name=req["TableName"],
            direction=req["Direction"],
            inclusive_start_primary_key=[(i["Key"], pk_val(i["Val"]) if "Val" in i else tablestore.INF_MIN) for i in req["StartPrimaryKey"]],
            exclusive_end_primary_key=[(i["Key"], pk_val(i["Val"]) if "Val" in i else tablestore.INF_MAX) for i in req["EndPrimaryKey"]],
            max_version=req["MaxVersion"],
        )

        return [dict([(i[0], i[1]) for i in row.primary_key]) | dict([(i[0], i[1]) for i in row.attribute_columns]) for row in rows]


def pk_val(val):
    if val == "INF_MAX":
        return tablestore.INF_MAX
    if val == "INF_MIN":
        return tablestore.INF_MIN
    return val





#!/usr/bin/env python3


import json

from .hook import Hook
from ..result import TestResult, CaseResult, StepResult
from ..util import merge


class DebugHook(Hook):
    def __init__(self, args=None):
        super().__init__(args)
        args = merge(args, {
            "padding": "  "
        })
        self.padding_to_add = args["padding"]
        self.padding = ""

    def on_test_start(self, test):
        print("{}{i18n.title.test} {name}".format(self.padding, name=test, i18n=self.i18n))
        self.padding += self.padding_to_add

    def on_test_end(self, res: TestResult):
        self.padding = self.padding[:-len(self.padding_to_add)]
        print("{}{i18n.title.test} {res.name}".format(self.padding, res=res, i18n=self.i18n))

    def on_case_start(self, case_info):
        print("{}{i18n.testHeader.case} {name}".format(self.padding, name=case_info["name"], i18n=self.i18n))
        self.padding += self.padding_to_add
        DebugHook.debug_object(self.padding, "CaseInfo", case_info)

    def on_case_end(self, res: CaseResult):
        self.padding = self.padding[:-len(self.padding_to_add)]
        print("{}{i18n.testHeader.case} {res.name}".format(self.padding, res=res, i18n=self.i18n))

    def on_set_up_start(self, case_info):
        print("{}{i18n.testHeader.setUp} {name}".format(self.padding, name=case_info["name"], i18n=self.i18n))
        self.padding += self.padding_to_add
        DebugHook.debug_object(self.padding, "SetUpInfo", case_info)

    def on_set_up_end(self, res: CaseResult):
        self.padding = self.padding[:-len(self.padding_to_add)]
        print("{}{i18n.testHeader.setUp} {res.name}".format(self.padding, res=res, i18n=self.i18n))

    def on_tear_down_start(self, case_info):
        print("{}{i18n.testHeader.tearDown} {name}".format(self.padding, name=case_info["name"], i18n=self.i18n))
        self.padding += self.padding_to_add
        DebugHook.debug_object(self.padding, "TearDownInfo", case_info)

    def on_tear_down_end(self, res: CaseResult):
        self.padding = self.padding[:-len(self.padding_to_add)]
        print("{}{i18n.testHeader.tearDown} {res.name}".format(self.padding, res=res, i18n=self.i18n))

    def on_step_start(self, step_info):
        print("{}{i18n.caseHeader.step} {name}".format(self.padding, name=step_info["name"], i18n=self.i18n))
        self.padding += self.padding_to_add
        DebugHook.debug_object(self.padding, "Step", step_info)

    def on_step_end(self, res: StepResult):
        self.padding = self.padding[:-len(self.padding_to_add)]
        print("{}{i18n.caseHeader.step} {res.name}".format(self.padding, res=res, i18n=self.i18n))

    @staticmethod
    def debug_object(padding, title, result):
        print("\n".join([padding + line for line in ("{}: {}".format(title, json.dumps(result, indent=2))).split("\n")]))



#!/usr/bin/env python3


from ..result import TestResult, CaseResult, StepResult
from ..i18n import I18n


class Hook:
    def __init__(self, args=None):
        self.i18n = I18n(args).i18n()

    def on_exit(self, res: TestResult):
        pass

    def on_test_start(self, test):
        pass

    def on_test_end(self, res: TestResult):
        pass

    def on_case_start(self, case_info):
        pass

    def on_case_end(self, res: CaseResult):
        pass

    def on_set_up_start(self, case_info):
        pass

    def on_set_up_end(self, res: CaseResult):
        pass

    def on_tear_down_start(self, case_info):
        pass

    def on_tear_down_end(self, res: CaseResult):
        pass

    def on_step_start(self, step_info):
        pass

    def on_step_end(self, res: StepResult):
        pass



#!/usr/bin/env python3


import datetime
import json
import unittest
from aliyun.log import LogClient, PutLogsRequest, LogItem, GetLogsRequest

from .sls_driver import SLSDriver


class TestSLS(unittest.TestCase):
    def test_sls(self):
        client = LogClient(
            "cn-shanghai.log.aliyuncs.com",
            "xx",
            "xx",
        )

        res = client.put_logs(PutLogsRequest(
            project="imm-dev-hl",
            logstore="convert-server-log",
            topic="test-topic",
            source="test-source",
            hashKey="123123",
            logtags=[("tag1", "val1"), ("tag2", "val2")],
            logitems=[
                LogItem(
                    timestamp=int(datetime.datetime.now().timestamp()),
                    contents=[
                        ("key1", "val1"),
                        ("key2", "val2"),
                    ],
                ),
                LogItem(
                    timestamp=int(datetime.datetime.now().timestamp()),
                    contents=[
                        ("key3", "val3"),
                        ("key4", "val4"),
                    ],
                )
            ],
        ))
        print(res)

        res = client.get_logs(GetLogsRequest(
            project="imm-dev-hl",
            logstore="convert-server-log",
            fromTime=int(datetime.datetime.now().timestamp()) - 3600,
            toTime=int(datetime.datetime.now().timestamp()),
            topic="test-topic",
        ))
        print(res)


class TestSLSDriver(unittest.TestCase):
    def test_sls_driver(self):
        driver = SLSDriver({
            "Endpoint": "cn-shanghai.log.aliyuncs.com",
            "AccessKeyId": "xx",
            "AccessKeySecret": "xx",
        })

        res = driver.do({
            "Action": "PutLogs",
            "Project": "imm-dev-hl",
            "Logstore": "convert-server-log",
            "Topic": "test-topic",
            "Source": "test-source",
            "HashKey": "123123",
            "LogTags": {
                "tag1": "val1",
                "tag2": "val2",
            },
            "LogItems": [{
                "Timestamp": int(datetime.datetime.now().timestamp()),
                "Contents": {
                    "key1": "val1",
                    "key2": "val2",
                    "key3": "val3",
                }
            }],
        })
        print(json.dumps(res, indent=2))

        res = driver.do({
            "Action": "GetLogs",
            "Project": "imm-dev-hl",
            "Logstore": "convert-server-log",
            "FromTime": int(datetime.datetime.now().timestamp()) - 3600,
            "ToTime": int(datetime.datetime.now().timestamp()),
            "Topic": "test-topic"
        })
        print(json.dumps(res, indent=2))


def main():
    unittest.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3


import json
import unittest

import bson.json_util
import pymongo
from bson.objectid import ObjectId

from .mongo_driver import MongoDriver, flat


class TestMongo(unittest.TestCase):
    def setUp(self) -> None:
        self.client = pymongo.MongoClient(
            host="localhost",
            port=27017,
        )
        self.database = self.client.get_database("some_database")
        self.collection = self.database.get_collection("some_collection")

    def test_count_documents(self):
        db = self.client.some_database
        col = db["some_collection"]
        total_docs = col.count_documents({})
        print(total_docs)

    def test_count_documents2(self):
        db = self.client.get_database("some_database")
        col = db.get_collection("some_collection")
        total_docs = col.count_documents({})
        print(total_docs)

    def test_insert_documents(self):
        # res = self.collection.insert_one({
        #   "key1": "value1",
        #   "key2": "value2",
        #   "key3": {
        #     "key4": "value5",
        #     "key5": 5,
        #     "key6": [
        #       "value61",
        #       "value62"
        #     ],
        #     "key7": [
        #       {
        #         "key8": "value8"
        #       }
        #     ]
        #   }
        # })

        res = self.collection.update_one({"_id": ObjectId("61ef739496089fa658380996")}, {
            "$set": {
                "key1": "value1",
                "key2": "value2",
                "key3": {
                    "key4": "value5",
                    "key5": 5,
                    "key6": [
                        "value61",
                        "value62"
                    ],
                    "key7": [
                        {
                            "key8": "value8"
                        }
                    ]
                }
            }
        })
        print(res)

    def test_get_documents(self):
        res = self.collection.find_one(ObjectId("61ef739496089fa658380996"))
        print(res)
        print(dict(res))
        print(json.dumps(dict(res), default=lambda x: str(x)))
        print(bson.json_util.dumps(res))

    def test_update_documents(self):
        res = self.collection.update_one({"_id": ObjectId("61ef739496089fa658380996")}, {"$set": {"key2": "val2"}}, upsert=True)
        print(res)
        print(res.matched_count, res.modified_count, res.acknowledged, res.upserted_id)


class TestMongoDriver(unittest.TestCase):
    def setUp(self) -> None:
        self.driver = MongoDriver({
            "host": "localhost",
            "port": 27017,
            "database": "some_database"
        })

    def test_insert_one(self):
        res = self.driver.do({
            "cmd": "insertOne",
            "collection": "some_collection",
            "document": {
              "key1": "value1",
              "key2": "value2",
              "key3": {
                "key4": "value5",
                "key5": 5,
                "key6": [
                  "value61",
                  "value62"
                ],
                "key7": [
                  {
                    "key8": "value8"
                  }
                ]
              }
            }
        })
        print(json.dumps(res))

    def test_find_one(self):
        res = self.driver.do({
            "cmd": "findOne",
            "collection": "some_collection",
            "_id": "61ef858608289cdcca960b89",
        })
        print(json.dumps(res, indent=2))

    def test_update_one(self):
        res = self.driver.do({
            "cmd": "updateOne",
            "collection": "some_collection",
            "_id": "61ef858608289cdcca960b89",
            "update": {
                "key3": {
                    "key4": "val55",
                }
            }
        })
        print(json.dumps(res))


class TestFlat(unittest.TestCase):
    def test_flat(self):
        d = flat({
          "key1": "value1",
          "key2": "value2",
          "key3": {
            "key4": "value5",
            "key5": 5,
            "key6": [
              "value61",
              "value62"
            ],
            "key7": [
              {
                "key8": "value8"
              }
            ]
          }
        })
        print(json.dumps(d, indent=2))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3


from .expect import *
import unittest


class TestExpectVal(unittest.TestCase):
    def test_expect_val(self):
        self.assertTrue(expect_eval("val % 2 == 1", val=1))
        self.assertTrue(expect_eval("val.endswith('v3')", val="eeb848f8611a4ff980d2e56a2760b4fcv3"))


class TestExpectObj(unittest.TestCase):
    def test_expect_dict_equal(self):
        res = expect({
            "status": 200,
            "json": {
                "hex": "533f6046eb7f610e",
                "num": "5998619086395760910",
            }
        }, {
            "status": 200,
            "json": {
                "hex": "533f6046eb7f610e",
                "num": "5998619086395760910",
            }
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0], ExpectResult(is_pass=True, message="OK", node="status", val=200, expect=200))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message="OK", node="json.hex", val="533f6046eb7f610e", expect="533f6046eb7f610e"))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message="OK", node="json.num", val="5998619086395760910", expect="5998619086395760910"))

    def test_expect_dict_not_equal(self):
        res = expect({
            "status": 200,
            "json": {
                "hex": "533f6046eb7f610e",
                "num": "5998619086395760910",
            }
        }, {
            "status": 201,
            "json": {
                "hex": "533f6046eb7f610e",
                "num": "xx",
            }
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message="NotEqual", node="status", val=200, expect=201))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message="OK", node="json.hex", val="533f6046eb7f610e", expect="533f6046eb7f610e"))
        self.assertEqual(res[2], ExpectResult(is_pass=False, message="NotEqual", node="json.num", val="5998619086395760910", expect="xx"))

    def test_expect_list_equal(self):
        res = expect({
            "status": 200,
            "json": [{
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }]
        }, {
            "status": 200,
            "json": [{
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }]
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[0], ExpectResult(is_pass=True, message="OK", node="status", val=200, expect=200))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message="OK", node="json.0.hex", val="111", expect="111"))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message="OK", node="json.0.num", val="222", expect="222"))
        self.assertEqual(res[3], ExpectResult(is_pass=True, message="OK", node="json.1.hex", val="333", expect="333"))
        self.assertEqual(res[4], ExpectResult(is_pass=True, message="OK", node="json.1.num", val="444", expect="444"))

    def test_expect_list_not_equal(self):
        res = expect({
            "status": 200,
            "json": [{
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }]
        }, {
            "status": 201,
            "json": [{
                "hex": "123",
                "num": "222",
            }, {
                "hex": "333",
                "num": "456",
            }]
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message="NotEqual", node="status", val=200, expect=201))
        self.assertEqual(res[1], ExpectResult(is_pass=False, message="NotEqual", node="json.0.hex", val="111", expect="123"))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message="OK", node="json.0.num", val="222", expect="222"))
        self.assertEqual(res[3], ExpectResult(is_pass=True, message="OK", node="json.1.hex", val="333", expect="333"))
        self.assertEqual(res[4], ExpectResult(is_pass=False, message="NotEqual", node="json.1.num", val="444", expect="456"))

    def test_expect_list_equal_2(self):
        res = expect([
            {
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }], [{
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }
        ])
        self.assertTrue(res)
        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], ExpectResult(is_pass=True, message="OK", node="0.hex", val="111", expect="111"))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message="OK", node="0.num", val="222", expect="222"))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message="OK", node="1.hex", val="333", expect="333"))
        self.assertEqual(res[3], ExpectResult(is_pass=True, message="OK", node="1.num", val="444", expect="444"))

    def test_expect_list_not_equal_2(self):
        res = expect([
            {
                "hex": "111",
                "num": "222",
            }, {
                "hex": "333",
                "num": "444",
            }], [{
                "hex": "123",
                "num": "222",
            }, {
                "hex": "333",
                "num": "456",
            }
        ])
        self.assertTrue(res)
        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message="NotEqual", node="0.hex", val="111", expect="123"))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message="OK", node="0.num", val="222", expect="222"))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message="OK", node="1.hex", val="333", expect="333"))
        self.assertEqual(res[3], ExpectResult(is_pass=False, message="NotEqual", node="1.num", val="444", expect="456"))

    def test_expect_dict_no_such_key_1(self):
        res = expect({
            "key1": "val1"
        }, {
            "key2": "val2"
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message='NoSuchKey', node='key2', val=None, expect='val2'))

    def test_expect_dict_no_such_key_2(self):
        res = expect({
            "key1": "val1"
        }, {
            "key2": {
                "key3": "val3",
                "key3": "val4"
            }
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message='NoSuchKey', node='key2', val=None, expect={'key3': 'val4'}))

    def test_expect_dict_no_such_key_3(self):
        res = expect({
            "key1": "val1"
        }, {
            "#key2": "val2"
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message='NoSuchKey', node='key2', val=None, expect='val2'))

    def test_expect_list_no_such_key(self):
        res = expect([1, 2], [1, 2, 3])
        self.assertTrue(res)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0], ExpectResult(is_pass=True, message='OK', node='0', val=1, expect=1))
        self.assertEqual(res[1], ExpectResult(is_pass=True, message='OK', node='1', val=2, expect=2))
        self.assertEqual(res[2], ExpectResult(is_pass=False, message='NoSuchKey', node='2', val=None, expect=3))

    def test_expect_type_diff_1(self):
        res = expect({
            "key1": [1, 2]
        }, {
            "key1": {"key2": "val2"}
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], ExpectResult(is_pass=False, message='TypeDiff', node='key1', val=[1, 2], expect={"key2": "val2"}))

    def test_expect_type_diff_2(self):
        res = expect({
            "json": [1, 2, 3]
        }, {
            "json": [1, {"key1": "val1"}, 3]
        })
        self.assertTrue(res)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0], ExpectResult(is_pass=True, message='OK', node='json.0', val=1, expect=1))
        self.assertEqual(res[1], ExpectResult(is_pass=False, message='TypeDiff', node='json.1', val=2, expect={"key1": "val1"}))
        self.assertEqual(res[2], ExpectResult(is_pass=True, message='OK', node='json.2', val=3, expect=3))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3


REQUIRED = "__DEFAULT__REQUIRED__"


def merge(req, dft):
    if req == None:
        return _merge_recursive("", {}, dft)
    return _merge_recursive("", req, dft)


def _merge_recursive(root: str, req, dft):
    if isinstance(dft, dict):
        for key, val in dft.items():
            root_dot_key = "{}.{}".format(root, key).lstrip(".")
            if isinstance(val, dict):
                req[key] = _merge_recursive(root_dot_key, req[key] if key in req and req[key] is not None else {}, val)
            elif isinstance(val, list):
                req[key] = _merge_recursive(root_dot_key, req[key] if key in req and req[key] is not None else [], val)
            else:
                if val == REQUIRED and key not in req:
                    raise Exception("missing required key [{}]".format(root_dot_key))
                elif key not in req:
                    req[key] = val
    if isinstance(dft, list):
        for idx, val in enumerate(dft):
            root_dot_key = "{}.{}".format(root, idx).lstrip(".")
            if isinstance(val, dict):
                while idx >= len(req):
                    req.append({})
                req[idx] = _merge_recursive(root_dot_key, req[idx], val)
            elif isinstance(val, list):
                while idx >= len(req):
                    req.append([])
                req[idx] = _merge_recursive(root_dot_key, req[idx], val)
            else:
                if val == REQUIRED and idx >= len(req):
                    raise Exception("missing required key [{}]".format(root_dot_key))
                elif idx >= len(req):
                    req.append(val)
    return req



#!/usr/bin/env python3


import unittest
import os


class TestTravel(unittest.TestCase):
    def test_travel(self):
        print([i for i in os.listdir("..") if os.path.isdir(os.path.join("..", i))])


class TestForLoop(unittest.TestCase):
    def test_for_loop(self):
        for i in range(1):
            print("body")
            break
        else:
            print("else")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3


from ..result import TestResult
from ..i18n import I18n


class Reporter:
    def __init__(self, args=None):
        self.i18n = I18n(args).i18n()

    def report(self, res: TestResult) -> str:
        return ""



#!/usr/bin/env python3


import os
import sys
import unittest
import pathlib
import importlib
import grpc
from .grpc_driver import GRPCDriver


# 启动服务参考 https://grpc.io/docs/languages/python/basics/
# 1. 下载代码： git clone -d 1 git@github.com:grpc/grpc.git
# 2. 安装工具: pip3 install grpcio-tools
# 3. 启动服务: cd grpc/examples/python/route_guide && python3 route_guide_server.py


class TestGRPC(unittest.TestCase):
    def test_hello(self):
        a = "../ops/grpc/route_guide.proto"
        p = pathlib.Path(a)
        print(p.parent.absolute())
        print(p.stem)

    def test_grpc(self):
        p = pathlib.Path("../ops/grpc/route_guide.proto")
        sys.path.append(str(p.parent.absolute()))
        name = "route_guide"
        pb2 = importlib.import_module(name + "_pb2")
        pb2_grpc = importlib.import_module(name + "_pb2_grpc")

        with grpc.insecure_channel('localhost:50051') as channel:
            stub = getattr(pb2_grpc, "RouteGuideStub")(channel)
            print((getattr(stub, "GetFeature"))(getattr(pb2, "Point")(**{
                "latitude": 409146138,
                "longitude": 746188906,
            })))


class TestGRPCDriver(unittest.TestCase):
    def test_grpc_driver(self):
        print(os.getcwd())
        d = GRPCDriver(args={
            "proto": "../ops/grpc/route_guide.proto",
            "endpoint": "localhost:50051",
            "service": "RouteGuideStub",
        })

        res = d.do({
            "method": "GetFeature",
            "type": "Point",
            "args": {
                "latitude": 409146138,
                "longitude": 746188906,
            }
        })

        print(res)


def main():
    unittest.main()


if __name__ == "__main__":
    main()




#!/usr/bin/env python3


from dataclasses import dataclass
from datetime import timedelta


@dataclass
class ExpectResult:
    is_pass: bool
    message: str
    node: str
    val: any
    expect: any

    def to_json(self):
        return {
            "isPass": self.is_pass,
            "message": self.message,
            "node": self.node,
            "val": self.val,
            "expect": self.expect,
        }

    @staticmethod
    def from_json(obj):
        res = ExpectResult(
            is_pass=obj["isPass"],
            message=obj["message"],
            node=obj["node"],
            val=obj["val"],
            expect=obj["expect"],
        )
        return res

@dataclass
class SubStepResult:
    is_pass: bool
    is_err: bool
    err: str
    req: dict
    res: dict
    assertions: list[ExpectResult]
    assertion_pass: int
    assertion_fail: int
    elapse: timedelta

    def to_json(self):
        return {
            "isPass": self.is_pass,
            "isErr": self.is_err,
            "err": self.err,
            "req": self.req,
            "res": self.res,
            "assertions":  self.assertions,
            "assertionPass": self.assertion_pass,
            "assertionFail": self.assertion_fail,
            "elapse": int(self.elapse.total_seconds() * 1000000),
        }

    @staticmethod
    def from_json(obj):
        res = SubStepResult()
        res.is_pass = obj["isPass"]
        res.is_err = obj["isErr"]
        res.err = obj["err"]
        res.req = obj["req"]
        res.res = obj["res"]
        res.assertions = [ExpectResult.from_json(i) for i in obj["assertions"]]
        res.assertion_pass = obj["assertionPass"]
        res.assertion_fail = obj["assertionFail"]
        res.elapse = timedelta(microseconds=obj["elapse"])
        return res

    def __init__(self):
        self.is_pass = True
        self.is_err = False
        self.err = ""
        self.req = {}
        self.res = {}
        self.assertions = list[ExpectResult]()
        self.assertion_pass = 0
        self.assertion_fail = 0
        self.elapse = timedelta(seconds=0)

    def set_error(self, message):
        self.is_pass = False
        self.is_err = True
        self.err = message
        self.assertion_fail += 1

    def add_expect_result(self, result):
        self.assertions = result
        self.assertion_pass += sum(1 for i in self.assertions if i.is_pass)
        self.assertion_fail += len(self.assertions) - self.assertion_pass
        self.is_pass = self.assertion_fail == 0


@dataclass
class StepResult:
    name: str
    description: str
    ctx: str
    is_skip: bool
    is_pass: bool
    req: dict
    res: dict
    sub_steps: list[SubStepResult]
    assertion_pass: int
    assertion_fail: int
    elapse: timedelta

    def to_json(self):
        return {
            "name": self.name,
            "description": self.description,
            "ctx": self.ctx,
            "isSkip": self.is_skip,
            "isPass": self.is_pass,
            "req": self.req,
            "res": self.res,
            "subSteps": self.sub_steps,
            "assertionPass": self.assertion_pass,
            "assertionFail": self.assertion_fail,
            "elapse": int(self.elapse.total_seconds() * 1000000),
        }

    @staticmethod
    def from_json(obj):
        res = StepResult(obj["name"], obj["ctx"], description=obj["description"])
        res.is_skip = obj["isSkip"]
        res.is_pass = obj["isPass"]
        res.req = obj["req"]
        res.res = obj["res"]
        res.sub_steps = [SubStepResult.from_json(i) for i in obj["subSteps"]]
        res.assertion_pass = obj["assertionPass"]
        res.assertion_fail = obj["assertionFail"]
        res.elapse = timedelta(microseconds=obj["elapse"])
        return res

    def __init__(self, name, ctx, description="", is_skip=False):
        self.name = name
        self.ctx = ctx
        self.description = description
        self.is_skip = is_skip
        self.is_pass = True
        self.req = {}
        self.res = {}
        self.sub_steps = list[SubStepResult]()
        self.assertion_pass = 0
        self.assertion_fail = 0
        self.elapse = timedelta(seconds=0)

    def add_sub_step_result(self, result: SubStepResult):
        self.req = result.req
        self.res = result.res
        self.sub_steps.append(result)
        self.assertion_pass += result.assertion_pass
        self.assertion_fail += result.assertion_fail
        self.is_pass = self.assertion_fail == 0


@dataclass
class CaseResult:
    id_: str
    name: str
    directory: str
    description: str
    command: str
    status: str
    before_case_steps: list[StepResult]
    pre_steps: list[StepResult]
    steps: list[StepResult]
    post_steps: list[StepResult]
    after_case_steps: list[StepResult]
    is_pass: bool
    is_skip: bool
    step_pass: int
    step_fail: int
    step_skip: int
    assertion_pass: int
    assertion_fail: int
    elapse: timedelta

    def to_json(self):
        return {
            "id": self.id_,
            "name": self.name,
            "directory": self.directory,
            "description": self.description,
            "command": self.command,
            "status": self.status,
            "elapse": int(self.elapse.total_seconds() * 1000000),
            "isPass": self.is_pass,
            "isSkip": self.is_skip,
            "steps": self.steps,
            "beforeCaseSteps": self.before_case_steps,
            "afterCaseSteps": self.after_case_steps,
            "stepPass": self.step_pass,
            "stepFail": self.step_fail,
            "assertionPass": self.assertion_pass,
            "assertionFail": self.assertion_fail,
        }

    @staticmethod
    def from_json(obj):
        res = CaseResult(directory=obj["directory"], id_=obj["id"], name=obj["name"], description=obj["description"])
        res.command = obj["command"]
        res.status = obj["status"]
        res.is_skip = obj["isSkip"]
        res.is_pass = obj["isPass"]
        res.before_case_steps = [StepResult.from_json(i) for i in obj["beforeCaseSteps"]]
        res.steps = [StepResult.from_json(i) for i in obj["steps"]]
        res.after_case_steps = [StepResult.from_json(i) for i in obj["afterCaseSteps"]]
        res.step_pass = obj["stepPass"]
        res.step_fail = obj["stepFail"]
        res.assertion_pass = obj["assertionPass"]
        res.assertion_fail = obj["assertionFail"]
        res.elapse = timedelta(microseconds=obj["elapse"])
        return res

    def __init__(self, directory, id_, name, description="", command="", is_skip=False):
        self.id_ = id_
        self.name = name
        self.directory = directory
        self.description = description
        self.command = command
        self.status = "pass"
        self.is_skip = is_skip
        self.before_case_steps = list[StepResult]()
        self.pre_steps = list[StepResult]()
        self.steps = list[StepResult]()
        self.post_steps = list[StepResult]()
        self.after_case_steps = list[StepResult]()
        self.is_pass = True
        self.elapse = timedelta(seconds=0)
        self.assertion_pass = 0
        self.assertion_fail = 0
        self.step_pass = 0
        self.step_fail = 0
        self.step_skip = 0
        if is_skip:
            self.status = "skip"

    def add_case_step_result(self, step: StepResult):
        self.steps.append(step)
        if step.is_skip:
            self.step_skip += 1
        elif not step.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.step_fail += 1
        else:
            self.step_pass += 1
        self.assertion_pass += step.assertion_pass
        self.assertion_fail += step.assertion_fail

    def add_case_pre_step_result(self, step: StepResult):
        self.pre_steps.append(step)
        if step.is_skip:
            self.step_skip += 1
        elif not step.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.step_fail += 1
        else:
            self.step_pass += 1
        self.assertion_pass += step.assertion_pass
        self.assertion_fail += step.assertion_fail

    def add_case_post_step_result(self, step: StepResult):
        self.post_steps.append(step)
        if step.is_skip:
            self.step_skip += 1
        elif not step.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.step_fail += 1
        else:
            self.step_pass += 1
        self.assertion_pass += step.assertion_pass
        self.assertion_fail += step.assertion_fail

    def add_before_case_step_result(self, step: StepResult):
        self.before_case_steps.append(step)
        if not step.is_pass:
            self.is_pass = False
            self.status = "fail"

    def add_after_case_step_result(self, step: StepResult):
        self.after_case_steps.append(step)
        if not step.is_pass:
            self.is_pass = False
            self.status = "fail"

@dataclass
class TestResult:
    directory: str
    name: str
    description: str
    status: str
    is_skip: bool
    is_pass: bool
    is_err: bool
    err: str
    set_ups: list[CaseResult]
    cases: list[CaseResult]
    tear_downs: list[CaseResult]
    sub_tests: list
    elapse: timedelta
    sub_tests: list
    case_pass: int
    case_fail: int
    case_skip: int
    set_up_pass: int
    set_up_fail: int
    tear_down_pass: int
    tear_down_fail: int
    step_pass: int
    step_fail: int
    step_skip: int
    assertion_pass: int
    assertion_fail: int
    curr_case_pass: int
    curr_case_fail: int
    curr_case_skip: int
    sub_test_pass: int
    sub_test_skip: int
    sub_test_fail: int

    def to_json(self):
        return {
            "directory": self.directory,
            "name": self.name,
            "status": self.status,
            "description": self.description,
            "isSkip": self.is_skip,
            "isPass": self.is_pass,
            "isErr": self.is_err,
            "err": self.err,
            "elapse": int(self.elapse.total_seconds() * 1000000),
            "casePass": self.case_pass,
            "caseFail": self.case_fail,
            "caseSkip": self.case_skip,
            "setUpPass": self.set_up_pass,
            "setUpFail": self.set_up_fail,
            "tearDownPass": self.tear_down_pass,
            "tearDownFail": self.tear_down_fail,
            "currCasePass": self.curr_case_pass,
            "currCaseFail": self.curr_case_fail,
            "currCaseSkip": self.curr_case_skip,
            "stepPass": self.step_pass,
            "stepFail": self.step_fail,
            "stepSkip": self.step_skip,
            "assertionPass": self.assertion_pass,
            "assertionFail": self.assertion_fail,
            "subTestPass": self.sub_test_pass,
            "subTestSkip": self.sub_test_skip,
            "subTestFail": self.sub_test_fail,
            "cases": self.cases,
            "setUps": self.set_ups,
            "tearDowns": self.tear_downs,
            "subTests": self.sub_tests,
        }

    @staticmethod
    def from_json(obj):
        res = TestResult(directory=obj["directory"], name=obj["name"], description=obj["description"], err_message=obj["err"], is_skip=obj["isSkip"])
        res.is_pass = obj["isPass"]
        res.status = obj["status"]
        res.set_ups = [CaseResult.from_json(i) for i in obj["setUps"]]
        res.cases = [CaseResult.from_json(i) for i in obj["cases"]]
        res.tear_downs = [CaseResult.from_json(i) for i in obj["tearDowns"]]
        res.sub_tests = [TestResult.from_json(i) for i in obj["subTests"]]
        res.case_pass = obj["casePass"]
        res.case_fail = obj["caseFail"]
        res.case_skip = obj["caseSkip"]
        res.set_up_pass = obj["setUpPass"]
        res.set_up_fail = obj["setUpFail"]
        res.tear_down_pass = obj["tearDownPass"]
        res.tear_down_fail = obj["tearDownFail"]
        res.curr_case_pass = obj["currCasePass"]
        res.curr_case_fail = obj["currCaseFail"]
        res.curr_case_skip = obj["currCaseSkip"]
        res.step_pass = obj["stepPass"]
        res.step_fail = obj["stepFail"]
        res.step_skip = obj["stepSkip"]
        res.assertion_pass = obj["assertionPass"]
        res.assertion_fail = obj["assertionFail"]
        res.sub_test_pass = obj["subTestPass"]
        res.sub_test_skip = obj["subTestSkip"]
        res.sub_test_fail = obj["subTestFail"]
        res.elapse = timedelta(microseconds=obj["elapse"])
        return res

    def __init__(self, directory, name, description="", err_message=None, is_skip=False):
        self.directory = directory
        self.name = name
        self.description = description
        self.is_skip = is_skip
        self.is_pass = True
        self.is_err = False
        self.err = ""
        self.set_ups = list[CaseResult]()
        self.cases = list[CaseResult]()
        self.tear_downs = list[CaseResult]()
        self.sub_tests = list[TestResult]()
        self.elapse = timedelta(seconds=0)
        self.case_pass = 0
        self.case_fail = 0
        self.case_skip = 0
        self.set_up_pass = 0
        self.set_up_fail = 0
        self.tear_down_pass = 0
        self.tear_down_fail = 0
        self.curr_case_pass = 0
        self.curr_case_fail = 0
        self.curr_case_skip = 0
        self.assertion_pass = 0
        self.assertion_fail = 0
        self.step_pass = 0
        self.step_fail = 0
        self.step_skip = 0
        self.sub_test_pass = 0
        self.sub_test_skip = 0
        self.sub_test_fail = 0
        if err_message:
            self.is_pass = False
            self.is_err = True
            self.err = err_message
            self.case_fail += 1
        if self.is_skip:
            self.status = "skip"
        elif self.is_pass:
            self.status = "pass"
        else:
            self.status = "fail"

    def add_set_up_result(self, case: CaseResult):
        self.set_ups.append(case)
        if not case.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.set_up_fail += 1
        else:
            self.set_up_pass += 1

    def add_case_result(self, case: CaseResult):
        self.cases.append(case)
        if case.is_skip:
            self.case_skip += 1
            self.curr_case_skip += 1
        elif not case.is_pass:
            self.case_fail += 1
            self.is_pass = False
            self.status = "fail"
            self.curr_case_fail += 1
        else:
            self.case_pass += 1
            self.curr_case_pass += 1
        self.step_pass += case.step_pass
        self.step_fail += case.step_fail
        self.step_skip += case.step_skip
        self.assertion_pass += case.assertion_pass
        self.assertion_fail += case.assertion_fail

    def add_tear_down_result(self, case):
        self.tear_downs.append(case)
        if not case.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.tear_down_fail += 1
        else:
            self.tear_down_pass += 1

    def add_sub_test_result(self, sub_test):
        if sub_test.case_pass + sub_test.case_skip + sub_test.case_fail == 0:
            return

        self.sub_tests.append(sub_test)
        if sub_test.is_skip:
            self.sub_test_skip += 1
            return
        self.case_pass += sub_test.case_pass
        self.case_fail += sub_test.case_fail
        self.case_skip += sub_test.case_skip
        self.step_pass += sub_test.step_pass
        self.step_fail += sub_test.step_fail
        self.step_skip += sub_test.step_skip
        self.assertion_pass += sub_test.assertion_pass
        self.assertion_fail += sub_test.assertion_fail
        if not sub_test.is_pass:
            self.is_pass = False
            self.status = "fail"
            self.sub_test_fail += 1
        else:
            self.sub_test_pass += 1



#!/usr/bin/env python3


import unittest
from .merge import *


class TestMerge(unittest.TestCase):
    def test_merge_dict(self):
        req = merge({
            "key1": "val1",
            "key2": {
                "key3": "val3",
            },
            "key6": [{
                "key7": "val7",
            }, {
                "key8": "val8",
            }]
        }, {
            "key2": {
                "key3": "dftVal3",
                "key4": "dftVal4",
            },
            "key5": "dftVal5",
            "key6": [{
                "key7": "dftVal7",
                "key8": "dftVal8",
            }, {
                "key7": "dftVal7",
                "key8": "dftVal8",
            }]
        })

        self.assertDictEqual(req, {
            "key1": "val1",
            "key2": {
                "key3": "val3",
                "key4": "dftVal4"
            },
            "key5": "dftVal5",
            "key6": [{
                "key7": "val7",
                "key8": "dftVal8",
            }, {
                "key7": "dftVal7",
                "key8": "val8",
            }]
        })

    def test_merge_list(self):
        req = merge([{
            "key7": "val7",
        }, {
            "key8": "val8",
        }], [{
            "key7": "dftVal7",
            "key8": "dftVal8",
        }, {
            "key7": "dftVal7",
            "key8": "dftVal8",
        }])
        self.assertListEqual(req, [{
            "key7": "val7",
            "key8": "dftVal8",
        }, {
            "key7": "dftVal7",
            "key8": "val8",
        }])

    def test_merge_required(self):
        with self.assertRaises(Exception) as cm:
            merge({
                "key1": "val1",
                "key2": {
                    "key3": "val3"
                }
            }, {
                "key1": REQUIRED,
                "key2": {
                    "key4": REQUIRED,
                }
            })
        self.assertTrue("missing required key [key2.key4]" in str(cm.exception))

    def test_merge_ots_create_table(self):
        dft = {
            "TableMeta": {
                "TableName": REQUIRED,
                "SchemeEntry": [[REQUIRED, REQUIRED]]
            },
            "TableOptions": {
                "TimeToLive": -1,
                "MaxVersion": 1,
                "MaxTimeDeviation": 86400,
            }
        }

        with self.assertRaises(Exception) as cm:
            merge({}, dft)
        self.assertTrue("missing required key [TableMeta.TableName]" in str(cm.exception))

        with self.assertRaises(Exception) as cm:
            merge({
                "TableMeta": {
                    "TableName": "test-table",
                },
            }, dft)
        self.assertTrue("missing required key [TableMeta.SchemeEntry.0.0]" in str(cm.exception))

        with self.assertRaises(Exception) as cm:
            merge({
                "TableMeta": {
                    "TableName": "test-table",
                    "SchemeEntry": [
                        ["key"]
                    ]
                },
            }, dft)
        self.assertTrue("missing required key [TableMeta.SchemeEntry.0.1]" in str(cm.exception))


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python3


from ..result import ExpectResult
from ..util.include import *
from ..util import exec_with_res


def expect(vals, rules, case=None, step=None, var=None, x=None, peval="#", pexec="%", mode=""):
    if mode == peval:
        return [run_expect("", rules, "eval", val=vals, case=case, step=step, var=var, x=x)]
    if mode == pexec:
        return [run_expect("", rules, "exec", val=vals, case=case, step=step, var=var, x=x)]

    if not rules:
        return []

    results = []
    _expect_recursive("", results, vals, rules, case=case, step=step, var=var, x=x, peval=peval, pexec=pexec)
    return results


def _expect_recursive(root: str, results: list[ExpectResult], vals, rules, case=None, step=None, var=None, x=None, peval="#", pexec="%"):
    if isinstance(rules, dict):
        for key, rule in rules.items():
            root_dot_key = "{}.{}".format(root, key.lstrip(peval).lstrip(pexec)).lstrip(".")
            if key.startswith(pexec):
                if key[len(pexec):] not in vals:
                    results.append(ExpectResult(is_pass=False, message="NoSuchKey", node=root_dot_key, val=None, expect=rule))
                else:
                    results.append(run_expect(root_dot_key, rule, "exec", val=vals[key[len(pexec):]], case=case, step=step, var=var, x=x))
            elif key.startswith(peval):
                if key[len(peval):] not in vals:
                    results.append(ExpectResult(is_pass=False, message="NoSuchKey", node=root_dot_key, val=None, expect=rule))
                else:
                    results.append(run_expect(root_dot_key, rule, "eval", val=vals[key[len(peval):]], case=case, step=step, var=var, x=x))
            elif key not in vals:
                results.append(ExpectResult(is_pass=False, message="NoSuchKey", node=root_dot_key, val=None, expect=rule))
            elif isinstance(rule, dict) or isinstance(rule, list):
                if not isinstance(vals[key], type(rule)):
                    results.append(ExpectResult(is_pass=False, message="TypeDiff", node=root_dot_key, val=vals[key], expect=rule))
                else:
                    _expect_recursive(root_dot_key, results, vals[key], rule, case=case, step=step, var=var, x=x, peval=peval, pexec=pexec)
            else:
                results.append(run_expect(root_dot_key, rule, "equal", val=vals[key], case=case, step=step, var=var, x=x))
    elif isinstance(rules, list):
        for idx, rule in enumerate(rules):
            root_dot_key = "{}.{}".format(root, idx).lstrip(".")
            if idx >= len(vals):
                results.append(ExpectResult(is_pass=False, message="NoSuchKey", node=root_dot_key, val=None, expect=rule))
            elif isinstance(rule, dict) or isinstance(rule, list):
                if not isinstance(vals[idx], type(rule)):
                    results.append(ExpectResult(is_pass=False, message="TypeDiff", node=root_dot_key, val=vals[idx], expect=rule))
                else:
                    _expect_recursive(root_dot_key, results, vals[idx], rule, case=case, step=step, var=var, x=x, peval=peval, pexec=pexec)
            else:
                results.append(run_expect(root_dot_key, rule, "equal", val=vals[idx], case=case, step=step, var=var, x=x))
    else:
        if vals == rules:
            results.append(ExpectResult(is_pass=True, message="OK", node=root, val=vals, expect=rules))
        else:
            results.append(ExpectResult(is_pass=False, message="NotEqual", node=root, val=vals, expect=rules))


def run_expect(root, rule, func, val=None, case=None, step=None, var=None, x=None):
    if func == "eval":
        ok, res = expect_eval(rule, val=val, case=case, step=step, var=var, x=x)
        if not ok:
            return ExpectResult(is_pass=False, message="EvalFail", node=root, val=val, expect="{} = {}".format(res, rule))
        return ExpectResult(is_pass=True, message="OK", node=root, val=val, expect=rule)
    elif func == "exec":
        ok, res = expect_exec(rule, val=val, case=case, step=step, var=var, x=x)
        if not ok:
            return ExpectResult(is_pass=False, message="ExecFail", node=root, val=val, expect="{} = {}".format(res, rule))
        return ExpectResult(is_pass=True, message="OK", node=root, val=val, expect=rule)
    else:
        if val != rule:
            return ExpectResult(is_pass=False, message="NotEqual", node=root, val=val, expect=rule)
        return ExpectResult(is_pass=True, message="OK", node=root, val=val, expect=rule)


def expect_eval(rule, val=None, case=None, step=None, var=None, x=None):
    res = eval(rule)
    if not isinstance(res, bool):
        return res == val, res
    return res, res


def expect_exec(rule, val=None, case=None, step=None, var=None, x=None):
    res = exec_with_res(rule, val=val, case=case, step=step, var=var, x=x)
    if not isinstance(res, bool):
        return res == val, res
    return res, res


def check(rule, val=None, case=None, step=None, var=None, x=None):
    ok, res = expect_eval(rule, val=val, case=case, step=step, var=var, x=x)
    return ok



#!/usr/bin/env python3


import json
import pathlib
import sys
import subprocess

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol
from thrift.TSerialization import serialize
from thrift.protocol.TJSONProtocol import TSimpleJSONProtocolFactory

from ..util import merge, REQUIRED
from .driver import Driver


class ThriftDriver(Driver):
    def __init__(self, args):
        args = merge(args, {
            "proto": REQUIRED,
            "endpoint": REQUIRED,
            "service": REQUIRED
        })
        p = pathlib.Path(args["proto"])
        self.proto_path = p
        self.generate_proto()
        sys.path.append("{}/gen-py".format(p.parent.absolute()))

        host, port = args["endpoint"].split(":")
        socket = TSocket.TSocket(host=host, port=int(port))
        transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(transport)
        transport.open()

        service = __import__("{}.{}".format(p.stem, args["service"]), fromlist=[p.stem])
        self.ttypes = __import__("{}.ttypes".format(p.stem), fromlist=[p.stem])
        self.client = getattr(service, "Client")(protocol)

    def generate_proto(self):
        command = "thrift -r --gen py {source}".format(source=str(self.proto_path.absolute()))
        process = subprocess.run(
            command.split(),
            cwd=str(self.proto_path.parent.absolute()),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if process.returncode != 0:
            raise Exception(process.stderr)

    def name(self, req):
        return req["method"]

    def do(self, req):
        req = merge(req, {
            "method": REQUIRED,
            "args": [],
        })

        args = []
        for arg in req["args"]:
            if isinstance(arg, dict):
                args.append(getattr(self.ttypes, arg["type"])(**arg["args"]))
            else:
                args.append(arg)
        res = getattr(self.client, req["method"])(*args)
        if hasattr(res, "write"):
            return json.loads(serialize(res, protocol_factory=TSimpleJSONProtocolFactory()))
        return res



#!/usr/bin/env python3

import unittest
import redis
from .redis_driver import *


# https://github.com/redis/redis-py
class TestRedis(unittest.TestCase):
    def setUp(self) -> None:
        self.client = redis.Redis(
            host='localhost',
            port=6379,
            db=0,
            password=None,
            decode_responses=True,
        )

    def test_set(self):
        ok = self.client.set("name", "hatlonely123")
        print(ok)

    def test_get(self):
        res = self.client.get("name")
        print(res)

    def test_hset(self):
        num = self.client.hset("key", "f1", "v1")
        print(num)

    def test_hget(self):
        res = self.client.hget("key", "f1")
        print(res)


class TestRedisDriver(unittest.TestCase):
    def setUp(self) -> None:
        self.driver = RedisDriver(args={})

    def test_set(self):
        res = self.driver.do(req={
            "cmd": "set",
            "key": "key1",
            "val": "val1",
            "exp": 60,
        })
        print(res)

    def test_get(self):
        res = self.driver.do(req={
            "cmd": "get",
            "key": "key1"
        })
        print(res)

    def test_set_json(self):
        res = self.driver.do(req={
            "cmd": "setJson",
            "key": "key1",
            "val": {
                "key1": "val1",
                "key2": "val2",
                "key3": 3,
            }
        })
        print(res)

    def test_get_json(self):
        res = self.driver.do(req={
            "cmd": "getJson",
            "key": "key1",
        })
        print(res)


if __name__ == '__main__':
    unittest.main()





#!/usr/bin/env python3


import unittest
from .mysql_driver import *


# https://github.com/PyMySQL/PyMySQL
class TestPymysql(unittest.TestCase):
    def setUp(self) -> None:
        self.connection = pymysql.connect(
            host='localhost',
            user='root',
            password='keaiduo1',
            database='hatlonely',
            cursorclass=pymysql.cursors.DictCursor,
        )

    def test_create_table(self):
        with self.connection.cursor() as cursor:
            sql = """
CREATE TABLE `users` (
`id` int(11) NOT NULL AUTO_INCREMENT,
`email` varchar(255) COLLATE utf8_bin NOT NULL,
`password` varchar(255) COLLATE utf8_bin NOT NULL,
PRIMARY KEY (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_bin
AUTO_INCREMENT=1;
"""
            cursor.execute(sql)
        self.connection.commit()

    def test_insert(self):
        with self.connection.cursor() as cursor:
            sql = "INSERT INTO `users` (`email`, `password`) VALUES (%s, %s)"
            cursor.execute(sql, ('webmaster@python.org', 'very-secret'))
        self.connection.commit()

    def test_select(self):
        with self.connection.cursor() as cursor:
            sql = "SELECT `id`, `password` FROM `users` WHERE `email`=%s"
            cursor.execute(sql, ('webmaster@python.org',))
            result = cursor.fetchall()
            print(result)
        self.connection.commit()


class TestMysqlDriver(unittest.TestCase):
    def setUp(self) -> None:
        self.driver = MysqlDriver(args={
            "host": "localhost",
            "port": 3306,
            "username": "root",
            "password": "keaiduo1",
            "database": "hatlonely",
        })

    def test_sql(self):
        res = self.driver.do(req={
            "cmd": "sql",
            "sql": "INSERT INTO `users` (`email`, `password`) VALUES (%s, %s)",
            "args": [
                "hatlonely@foxmail.com",
                "very-secret"
            ],
        })
        print(res)

    def test_fetchone(self):
        res = self.driver.do(req={
            "cmd": "fetchone",
            "sql": "SELECT `id`, `password` FROM `users` WHERE `email`=%s",
            "args": [
                "hatlonely@foxmail.com"
            ]
        })
        print(res)

    def test_fetchall(self):
        res = self.driver.do(req={
            "cmd": "fetchall",
            "sql": "SELECT `id`, `password` FROM `users` WHERE `email`=%s",
            "args": [
                "hatlonely@foxmail.com"
            ]
        })
        print(res)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3


import unittest
import requests

from .dingtalk_hook import *


class TestDingTalk(unittest.TestCase):
    def test_ding_talk(self):
        endpoint = "https://oapi.dingtalk.com/robot/send"
        access_token = "xxx"
        secret = "xxx"

        ts, sign = calculate_sign(secret)

        res = requests.post(
            url="https://oapi.dingtalk.com/robot/send",
            params={
                "access_token": access_token,
                "timestamp": ts,
                "sign": sign,
            },
            headers={
                "Content-Type": "application/json"
            },
            # json={
            #     "msgtype": "text",
            #     "text": {
            #         "content": "我就是我, 是不一样的烟火",
            #     }
            # },
            json={
                "msgtype": "actionCard",
                "actionCard": {
                    "title": "我 20 年前想打造一间苹果咖啡厅，而它正是 Apple Store 的前身",
                    "text": "![screenshot](https://img.alicdn.com/tfs/TB1NwmBEL9TBuNjy1zbXXXpepXa-2400-1218.png) \n\n #### 乔布斯 20 年前想打造的苹果咖啡厅 \n\n Apple Store 的设计正从原来满满的科技感走向生活化，而其生活化的走向其实可以追溯到 20 年前苹果一个建立咖啡馆的计划",
                    "btnOrientation": "0",
                    "btns": [
                        {
                            "title": "内容不错",
                            "actionURL": "https://www.dingtalk.com/"
                        },
                        {
                            "title": "不感兴趣",
                            "actionURL": "https://www.dingtalk.com/"
                        }
                    ]
                }
            }
        )
        print(res.status_code)
        print(res.text)

    def test_calculate_sign(self):
        ts, sign = calculate_sign("hello world")
        print(ts, sign)


class TestDingTalkHook(unittest.TestCase):
    def test_ding_talk(self):
        hook = DingTalkHook({
            "accessToken": "xxx",
            "secret": "xxx",
        })

        res = TestResult("", "", "")
        res.status = "pass"

        hook.on_exit(res)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3


import json
import traceback
import aliyunsdkcore.acs_exception.exceptions
from aliyunsdkcore.client import AcsClient
from aliyunsdkcore.request import CommonRequest

from ..util import merge, REQUIRED
from .driver import Driver


product_to_version = {
    "ots": "2016-06-20",
    "imm": "2017-09-06",
    "ram": "2015-05-01",
    "slb": "2014-05-15",
    "ack": "2015-12-15",
    "rds": "2014-08-15",
    "redis": "2015-01-01",
    "ecs": "2014-05-26",
    "vpc": "2016-04-28",
    "kms": "2016-01-20",
    "sts": "2015-04-01",
    "immv2": "2020-09-30"
}


class POPDriver(Driver):
    client: AcsClient
    endpoint: str
    product_id: str
    method: str
    scheme: str

    def __init__(self, args: dict):
        args = merge(args, {
            "AccessKeyId": REQUIRED,
            "AccessKeySecret": REQUIRED,
            "DisableVerify": False,
            "RegionId": "",
        })

        self.endpoint = args["Endpoint"].rstrip("/")
        self.client = AcsClient(args["AccessKeyId"], args["AccessKeySecret"], args["RegionId"], verify=not args["DisableVerify"])
        self.region_id = args["RegionId"]

    def name(self, req):
        return req["Action"]

    def do(self, req: dict):
        req = merge(req, {
            "Action": REQUIRED,
            "Method": "POST",
            "Scheme": "https",
            "ProductId": "",
            "Version": "",
            "RegionId": self.region_id,
            "Endpoint": self.endpoint,
        })

        creq = CommonRequest()
        creq.set_accept_format("json")
        creq.set_method(req["Method"])
        creq.set_protocol_type(req["Scheme"])
        if req["Endpoint"]:
            creq.set_domain(req["Endpoint"])

        if req["Version"]:
            creq.set_version(req["Version"])
        elif req["ProductId"]:
            creq.set_version(product_to_version[req["ProductId"]])
        else:
            raise Exception("unknown Version")

        creq.set_action_name(req["Action"])

        for key in req:
            if key in ["Action", "Version", "ProductId", "Scheme", "Method", "Endpoint"]:
                continue
            creq.add_query_param(key, req[key])

        try:
            res = self.client.do_action_with_exception(creq)
            return json.loads(str(res, encoding='utf-8'))
        except aliyunsdkcore.acs_exception.exceptions.ClientException as e:
            return {
                "Code": e.error_code,
                "Message": e.message,
            }
        except aliyunsdkcore.acs_exception.exceptions.ServerException as e:
            return {
                "Status": e.http_status,
                "Code": e.error_code,
                "Message": e.message,
                "RequestId": e.request_id,
            }
        except Exception as e:
            return {
                "Exception": traceback.format_exc(),
            }
